package org.eclipse.swt.widgets;

import il.co.cortext.rui.remote.IRuiClientListener;
import il.co.cortext.rui.remote.IRuiConstants;
import il.co.cortext.rui.remote.RuiClientEvent;
import il.co.cortext.rui.remote.RuiServerCommand;
import il.co.cortext.rui.remote.RuiSession;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.HelpListener;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.TraverseListener;
import org.eclipse.swt.graphics.Color;

public abstract class Control extends Widget
	implements IRuiClientListener, IRuiConstants {

	protected String     proxyClassName;
	protected Composite  parent;

	private String  windowId, widgetId, sessionAttachId, tooltipText;
	private Color   backgroundColor, foregroundColor;
	private Object  layoutData;
	private boolean isEnabled = true;
	private boolean isVisible = true;

	Control() {}

	// widgets
	public Control(Composite parent, int style)
		{ this(parent, style, null); }

	Control(Composite parent, int style, String proxyClassName) {
		super(parent, style);
		this.parent = parent;
		this.proxyClassName = (proxyClassName == null)?
			getDefaultProxyClassName(): proxyClassName;
		String id = getSession().makeWidgetId();
		init(parent.getWindowId(), id, id);
		sendToClient(METHOD_CREATE, parent.addChild(this));
		setStyle();
	}

	public void dispose() {
		if (parent != null) parent.removeChild(this);
		releaseFromSession();
		sendToClient("destroy");
		super.dispose();
	}

	public Composite getParent() { return parent; }
	public Shell     getShell()  { return parent.getShell(); }

	protected void init
		(String windowId, String widgetId,  String sessionAttachId) {
		this.windowId        = windowId;
		this.widgetId        = widgetId;
		this.sessionAttachId = sessionAttachId;
		getSession().setListener(sessionAttachId, this);
	}

	protected void releaseFromSession() {
		releaseChildrenFromSession();
		getSession().removeListener(sessionAttachId);
	}

	protected void releaseChildrenFromSession() {}

// layout ---------------------------------------------------------------------

	public Object getLayoutData() { return layoutData; }

	public void setLayoutData(Object layoutData)
		{ this.layoutData = layoutData; }

	public void pack() { pack(true); }

	public void pack(boolean changed) {}

	protected void sendLayoutDataToClient(Map state) {
		for (Iterator i = state.keySet().iterator(); i.hasNext();) {
			String key = (String) i.next();
			sendToClient(METHOD_LAYOUTHINT, key, (String) state.get(key));
		}
	}

// style ----------------------------------------------------------------------

	public Color getBackground() { return backgroundColor; }
	public Color getForeground() { return foregroundColor; }

	public void setBackground(Color backgroundColor) {
		this.backgroundColor = backgroundColor;
		sendToClient(METHOD_STYLE,
			"backgroundColor", getColorAsString(backgroundColor)
		);
	}

	public void setForeground(Color foregroundColor) {
		this.foregroundColor = foregroundColor;
		sendToClient(METHOD_STYLE,
			"color", getColorAsString(foregroundColor)
		);
	}

	protected String getColorAsString(Color color) {
		String red   = Integer.toHexString(color.getRed());
		String green = Integer.toHexString(color.getGreen());
		String blue  = Integer.toHexString(color.getBlue());
		if (red.length()   == 1) red   = "0" + red;
		if (green.length() == 1) green = "0" + green;
		if (blue.length()  == 1) blue  = "0" + blue;
		return "#" + red + green + blue;
	}

// misc control features ------------------------------------------------------

	public String  getTooltipText() { return tooltipText; }
	public boolean getEnabled()     { return isEnabled; }
	public boolean getVisible()     { return isVisible; }

	public void setToolTipText (String tooltipText) {
		this.tooltipText = tooltipText;
		sendToClient(METHOD_TOOLTIP, tooltipText);
	}

	public void setEnabled(boolean isEnabled) {
		this.isEnabled = isEnabled;
		sendToClient(METHOD_DISABLED, isEnabled? "0": "1");
	}

	public void setVisible (boolean isVisible) {
		this.isVisible = isVisible;
		sendToClient(METHOD_STYLE, "display", isVisible? "": "none");
	}

// event handling ------------------------------------------------------------- 

	public void handleClientEvent(RuiClientEvent clientEvent) 
		throws Exception {
		Event event = translateEvent(clientEvent);
		notifyListeners(event.type, event);
	}

// UI proxy interface ---------------------------------------------------------

	protected void sendToClient(String methodName) {
		String[] params = {};
		sendToClient(methodName, params);
	}

	protected void sendToClient(String methodName, String param) {
		String[] params = {param};
		sendToClient(methodName, params);
	}

	protected void sendToClient
		(String methodName, String param1, String param2) {
		String[] params = {param1, param2};
		sendToClient(methodName, params);
	}

	protected void sendToClient(String methodName, String[] params) {
		String widgetId = getWidgetId();
		String windowId = getWindowId();
		assert widgetId != null && windowId != null && proxyClassName != null;
		RuiServerCommand command = new RuiServerCommand
			(proxyClassName, methodName, windowId, widgetId, params);
		sendToClient(command);
	}

	protected void sendToClient(RuiServerCommand command)
		{ getSession().addCommand(command); }

// state info -----------------------------------------------------------------

	public Display getDisplay()  { return getShell().getDisplay(); }
	public String  getWindowId() { return windowId; }
	public String  getWidgetId() { return widgetId; }

	protected RuiSession getSession() { return Display.getSession(); }

// testing --------------------------------------------------------------------

	public String toString() {
		return "[" +
			proxyClassName + ":" + getWindowId() + ":" + getWidgetId() +
		"]";
	}

	public String getProxyClassName() { return proxyClassName; }

// mock methods for SWT compatibility -----------------------------------------

	public void setRedraw(boolean redraw) {}
	public void redraw   ()               {}

	public void addControlListener      (ControlListener listener)    {}
	public void addFocusListener        (FocusListener listener)      {}
	public void addHelpListener         (HelpListener listener)       {}
	public void addKeyListener          (HelpListener listener)       {}
	public void addMouseListener        (MouseListener listener)      {}
	public void addMouseMoveListener    (MouseMoveListener listener)  {}
	public void addMouseTrackListener   (MouseTrackListener listener) {}
	public void addPaintListener        (PaintListener listener)      {}
	public void addTraverseListener     (TraverseListener listener)   {}
	public void removeControlListener   (ControlListener listener)    {}
	public void removeFocusListener     (FocusListener listener)      {}
	public void removeHelpListener      (HelpListener listener)       {}
	public void removeMouseListener     (MouseListener listener)      {}
	public void removeMouseMoveListener (MouseMoveListener listener)  {}
	public void removeMouseTrackListener(MouseTrackListener listener) {}
	public void removePaintListener     (PaintListener listener)      {}
	public void removeTraverseListener  (TraverseListener listener)   {}
		
// template methods -----------------------------------------------------------

	protected void setStyle() {}

	protected Event translateEvent(RuiClientEvent clientEvent)
		throws Exception {
		Event event = new Event();
		event.type  = SWT.NONE;
		return event;
	}

	protected abstract String getDefaultProxyClassName();

}
