package org.eclipse.swt.widgets.tests;

import java.util.Map;

import il.co.cortext.rui.remote.IRuiConstants;

import junit.framework.TestCase;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

public abstract class ControlTestCase extends TestCase {

	protected ControlTestHelper helper;
	protected Shell shell;
	protected Display display;

	private Control subjectControl;

	public ControlTestCase(String name) { super(name); }

	protected void setUp() throws Exception {
		helper         = new ControlTestHelper();
		display        = new Display();
		shell          = new Shell(display);
		subjectControl = makeSubject();
		flush();
	}
		
	protected void tearDown() { helper.tearDown(); }

// control tests --------------------------------------------------------------

	public void testCreate() throws Exception {
		subjectControl = makeSubject();
		assertBufferMatches(subjectControl, IRuiConstants.METHOD_CREATE, "E1");
	}

	public void testDispose() throws Exception {
		subjectControl.dispose();
		assertBuffer(subjectControl, IRuiConstants.METHOD_DESTROY, null);
	}

	public void testGetShell() throws Exception
		{ assertEquals(shell, subjectControl.getShell()); }

	public void testBackground() throws Exception {
		Color color = new Color(display, 255, 0, 1);
		subjectControl.setBackground(color);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_STYLE,
			"backgroundColor.#ff0001"
		);
		assertEquals(color, subjectControl.getBackground());
	}

	public void testForeground() throws Exception {
		Color color = new Color(display, 1, 255, 0);
		subjectControl.setForeground(color);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_STYLE,
			"color.#01ff00"
		);
		assertEquals(color, subjectControl.getForeground());
	}

	public void testSetTooltipText() throws Exception {
		subjectControl.setToolTipText("foo");
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_TOOLTIP,
			"foo"
		);
		assertEquals("foo", subjectControl.getTooltipText());
	}

	public void testSetEnabled() throws Exception {
		assertEquals(true, subjectControl.getEnabled());

		subjectControl.setEnabled(false);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_DISABLED,
			"1"
		);
		assertEquals(false, subjectControl.getEnabled());

		subjectControl.setEnabled(true);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_DISABLED,
			"0"
		);
		assertEquals(true, subjectControl.getEnabled());
	}

	public void testSetVisible() throws Exception {
		assertEquals(true, subjectControl.getVisible());

		subjectControl.setVisible(false);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_STYLE,
			"display.none"
		);

		subjectControl.setVisible(true);
		assertBuffer(
			getDelegate(subjectControl),
			IRuiConstants.METHOD_STYLE,
			"display."
		);
		assertEquals(true, subjectControl.getVisible());
	}

// testing helpers ------------------------------------------------------------

	protected Control makeSubject() throws Exception
		{ return makeSubject(shell, SWT.NONE); }

	protected Control makeSubject(Composite parent, int style)
		throws Exception {
		Object[] args = {parent, new Integer(style)};
		return makeSubject(args);
	}


	protected Control makeSubject(Object[] args)
		throws Exception {
		return (Control)
			getSubjectClass().getConstructors()[0].newInstance(args);
	}

	private String makeCommand(
		Control subject, String proxyClassName, String methodName, String params
	) {
		return
			(
				proxyClassName == null?
					subject.getProxyClassName(): proxyClassName
			) + "." + methodName + "." + getCommandId(subject) +
			(params == null? "": "." + params) + "\n";
	}

	private String makeCommand
		(Control subject, String methodName, String params) {
		return makeCommand(subject, null, methodName, params);
	}

	protected Control getSubject() { return subjectControl; }
	protected String  flush()      { return helper.flush(); }

	protected void fireEvent(Map event) throws Exception
		{ helper.fireEvent(event); }

	protected String getFailMessage(String command, String buffer) {
		return
			("command: " + command + ", not in buffer: " + buffer).
			replace('\n', ';');
	}

// assertionss ----------------------------------------------------------------

	protected void assertBuffer(
		Control subject, String proxyClassName, String methodName, String params
	) {
		String command = makeCommand
			(subject, proxyClassName, methodName, params);
		String buffer = flush();
		if (!command.equals(buffer)) fail(getFailMessage(command, buffer));
	}

	protected void assertBuffer
		(Control subject, String methodName, String params)
			{ assertBuffer(subject, null, methodName, params); }

	protected void assertBufferMatches
		(Control subject, String methodName, String params) {
		String command = makeCommand(subject, methodName, params);
		String buffer  = flush();
		if(buffer.indexOf(command) == -1) fail(getFailMessage(command, buffer));
	}

	protected void assertBufferMatches(Object[][] expected) {
		String buffer = flush();
		for (int i = 0; i < expected.length; i++) {
			Object[] row      = expected[i];
			Control subject   = (Control) row[0];
			String methodName = (String)  row[1];
			String params     = (String)  row[2];
			String command    = makeCommand(subject, methodName, params);
			if(buffer.indexOf(command) == -1)
				fail(getFailMessage(command, buffer));
		}
	}

// template methods -----------------------------------------------------------

	protected Control getDelegate(Control subject) { return subject; }

	protected String getCommandId(Control subject)
		{ return subject.getWindowId() + "." + subject.getWidgetId(); }

	protected abstract Class getSubjectClass();

}
