package Rui::Application::Base;

use Carp;
use Attribute::Protected;
use Attribute::Abstract;
use Rui::Widget::Factory;
use Rui::Remote::SessionManager;

use base 'Rui::Window';

# creating -----------------------------------------------------------------

sub init
{
	my ($self, %params) = @_;
	$self->{windows}       = {};
	$self->{widgetFactory} = $params{widgetFactory} ||
		Rui::Widget::Factory->instance;
	(
		$params{session} ||
		Rui::Remote::SessionManager->instance->getActiveSession
	)->setApplication($self);
	$self->SUPER::init(%params);
	$self->start;
}

sub destroy
{
	my ($self, $dontSendToClient) = @_;
	foreach my $window (values %{$self->{windows}}) {
		$window->destroy(1) unless $window eq $self;
	}
	delete $self->{windows};
	my $session = $self->{session};
	$self->SUPER::destroy($dontSendToClient);
	$self->destroySession($session);
}

sub destroySession : Protected
	{ Rui::Remote::SessionManager->instance->destroySession(pop) }

# accessing ----------------------------------------------------------------

sub getWidgetFactory { shift->{widgetFactory} }
sub registerWindow   { shift->{windows}->{shift()} = pop }
sub unregisterWindow { delete shift->{windows}->{shift()} }

# template methods ---------------------------------------------------------

sub start : Abstract;

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Application::Base - abstract application base class

=head1 SYNOPSIS

  # subclassing
  package HelloWorld;
  use base 'Rui::Application::Base';
  sub start { shift->Label(value => 'Hello World') }

  # using as a session
  $application = HelloWorld->new;
  # print the program that is a description of the user
  # interface of the initial application state
  print join "\n", $session->flushBuffer;
  $application->destroy;

=head1 SUPERCLASS

L<Rui::Window>

=head1 REQUIRES

L<Rui::Widget::Factory>, Rui::Remote::SessionManager

=head1 DESCRIPTION

Rui applications inherit from this class. An application is a window,
but there is only one per session.

=head1 SUBCLASSING

The only thing to override is C<start()>, where you create the UI.
