
package Rui::Event::Listenable;

use Carp;
use Attribute::Protected;
use WeakRef;
use Rui::Event;

# creating -----------------------------------------------------------------

my $count = 0;

sub new
{
	my ($thing, %params) = @_;
	my $self = bless ({}, ref($thing) || $thing);
#$count++;my $padded = sprintf("%-55s", "O ".ref $self);print "$padded$count\n";
	$self->init(%params);
	return $self;
}

sub DESTROY
{
	my $self = shift;
#$count--;my $padded = sprintf("%-55s", ". ". ref $self);print "$padded$count\n";
}

sub init : Protected
{
	my ($self, %params) = @_;
	$self->{events} = {};
	my %events = exists $params{events}? %{$params{events}}: ();
	$self->attach;
	$self->addListener($_, $events{$_}) foreach keys %events;
}

# listening ----------------------------------------------------------------

sub addListener
{
	my ($self, $eventName, $listener) = @_;
	my $events = $self->{events};
	$events->{$eventName} = [] unless exists $events->{$eventName};
	my $listeners = $events->{$eventName};
	my $lastIndex = (push @$listeners, $listener) - 1;
	if (ref $listener eq 'ARRAY') { # type 3 listener
		weaken $listeners->[$lastIndex]->[1];
	} elsif (ref $listener ne 'CODE') { # type 2 listener
		weaken $listeners->[$lastIndex];
	}
	return $self;
}

sub removeListener
{
	my ($self, $eventName, $listener) = @_;
	$self->cleanListenerListForEvent($eventName);
	my $allListeners = $self->{events};
	my $listeners = $allListeners->{$eventName};
	croak "listenable has no listeners for listener $eventName"
		unless $listeners;

	my $oldSize = @$listeners;
	foreach my $i (0..@$listeners - 1) {
		my $l = $listeners->[$i];
		if (
			(ref $l eq 'ARRAY' && $listener eq $l->[1]) ||
			$listener eq $l
		) {
			splice (@$listeners, $i, 1);
			delete $self->{events}->{$eventName} if $oldSize == 1;
			last;
		}
	}
	croak "Listener not found: [$eventName, $self, $listener]"
		if $oldSize == @$listeners;
	return $self;
}

sub hasListenerForEvent : Protected
{
	my ($self, $eventName) = @_;
	$self->cleanListenerListForEvent($eventName);
	return exists $self->{events}->{$eventName};
}

sub getEventNamesOfListeners : Protected
{
	my $self = shift;
	$self->cleanListenerList;
	my %eventNames;
	$eventNames{$_} = 1 foreach keys %{$self->{events}};
	return sort keys %eventNames;
}

# event handling -----------------------------------------------------------

sub fireEvent : Protected
{
	my ($self, $event) = @_;
	my $eventName = $event->name;
	$self->cleanListenerListForEvent($eventName);
	return unless exists $self->{events}->{$eventName};
	my @listeners = map {$_} @{$self->{events}->{$eventName}};
	foreach my $listener (@listeners) {
		my $clone = $event->clone;
		if (ref $listener eq 'CODE') {
			&$listener($clone);
		} elsif (ref $listener eq 'ARRAY') {
			my $methodName = $listener->[0];
			$listener->[1]->$methodName($clone);
		} else {
			my $methodName = 'handleEvent_'. $eventName;
			$listener->$methodName($clone);
		}
	}
}

# memory management --------------------------------------------------------

sub cleanListenerList: Private
{
	my $self = shift;
	$self->cleanListenerListForEvent($_) foreach keys %{$self->{events}};
	return $self;
}

sub cleanListenerListForEvent: Private
{
	my ($self, $eventName) = @_;
	croak "no event name" unless $eventName;
	my $events = $self->{events};
	return $self unless exists $events->{$eventName};
	$events->{$eventName} = [grep {
		$_? ref $_ eq 'ARRAY'? $_->[1]? 1: 0: 1: 0;
	} @{$events->{$eventName}}];
	delete $events->{$eventName} unless @{$events->{$eventName}}; 
}

# attach-detach ------------------------------------------------------------

sub attach : Protected
{
	my $self = shift;
	foreach my $listenableName (keys %{$self->getAttachments}) {
		$self->attachTo($listenableName);
	}
}

sub detach : Protected
{
	my $self = shift;
	foreach my $listenableName (keys %{$self->getAttachments}) {
		$self->detachFrom($listenableName);
	}
}

sub attachTo : Protected
{
	my ($self, $listenableName) = @_;
	my $listenable = $self->$listenableName;
	return unless ref($listenable) &&
		UNIVERSAL::isa($listenable, 'Rui::Event::Listenable');
	my $listeners = $self->getAttachments->{$listenableName};
	croak "no listenable called $listenableName for $self" unless $listeners;
	$self->{"listenable_$listenableName"} = $listenable;
	foreach my $eventName (keys %$listeners) {
		my $myEventName = $listeners->{$eventName};
		croak "no handler name for $listenableName:$eventName" unless $myEventName;
		next unless $self->isAttachOk($eventName, $listenableName, $listenable);
		my $handlerName = "handleEvent_${listenableName}_$myEventName";
		$listenable->addListener($eventName => [$handlerName => $self]);
	}
}

sub detachFrom : Protected
{
	my ($self, $listenableName) = @_;
	my $listenable = delete $self->{"listenable_$listenableName"};
	return unless ref($listenable) &&
		UNIVERSAL::isa($listenable, 'Rui::Event::Listenable');
	my $listeners = $self->getAttachments->{$listenableName};
	croak "no listenable called $listenableName for $self" unless $listeners;
	my @listenerKeys = keys %$listeners;
	foreach my $eventName (keys %$listeners) {
		next unless $self->isAttachOk($eventName, $listenableName, $listenable);
		$listenable->removeListener($eventName => $self);
	}
}

sub getAttachments : Protected { {} }

sub isAttachOk : Protected
{
	my ($eventName, $listenableName, $listenable) = @_;
	return 1;
}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Event::Listenable - base class for listenable objects

=head1 SYNOPSIS

  # subclassing
  package Car;
  use base qw(Rui::Event::Listenable);
  ...
  sub openFrontLeftDoor {
     ...
     # DoorOpened is an  event class  
     $self->fireEvent(DoorOpened->new(
        name   => 'FrontLeftDoorOpened',
        source => $self,
        time   => time,
     ));
  }
  ...

  # using
  $car = Car->new(events => {     
     FrontLeftDoorOpened => sub {print shift->time}, # 1st param is event
     FrontLeftDoorClosed => sub {print shift->time},
  });

  $car->addListener(FrontLeftDoorOpened => $listener = sub { 
     my $event = shift;
     print 'name:'. $event->name. ' source:'. $event->source;
  });
  $car->removeListener($listener);
 
  $car->addListener(FrontLeftDoorOpened => $listener = MyListener->new);
  $car->removeListener($listener);
 
  $car->addListener( FrontLeftDoorClosed =>
     [CalledWhenDoorOpened => $listener = MyListener->new]
  );
  $car->removeListener($listener);
  
  # listening
  package MyListener;
  ...
  sub handleEvent_FrontLeftDoorOpened {
     my ($self, $event) = @_;
     ...
  }
  sub CalledWhenDoorClosed {
     my ($self, $event) = @_;
     ...
  }

=head1 REQUIRES

L<Rui::Event>

=head1 DESCRIPTION

An implementation of the observer pattern. Listenable objects fire
events: different events are fired by different objects. For each event
a name and a class are defined. For example: A car class defines an
event with a name C<FrontLeftDoorOpened>, and an event class
C<CarSoftware::Event::DoorOpened>. The event name is used as the key
when adding and removing listeners, and as a hint when the listener
needs to find a listener method.

You interface with a listenable object by adding and removing listeners.
Listeners are called when an event is fired, with a L<Rui::Event> as the
only parameter.

A listener is a DWIMer and can be one of the following:

=over 4

=item 1

callback - code ref to be called

=item 2

object - the method C<handleEvent_eventName> will be called

=item 3

array ref with two elements- scalar method name, object

=back

When the listener is an object (listener type 2), the method name to be
called is computed from the event name by adding C<handleEvent_> in
front of the event name. For example: a car object will call the method
C<handleEvent_FrontLeftDoorOpened> on its listeners that are objects.

When the listener is an array ref (listener type 3), the method name
(1st element) is called on the object (2nd element). When removing this
type of listener, you do not remove the array ref but the listener
object, i.e. exactly like you remove a type 2 listeners.

=head1 SUBCLASSING

For implementing listenable functionality, all you need to do is call
C<< $self->fireEvent($event) >> to fire an event. You can also override
C<add/removeListener()> to modify events fired by your class.

A listenable also features support for creating listener classes, thus
making it possible to create a listenable that is also a listener (as
most are). These classes may listen to several events on several
listenables. They will want to attach to all of them when created, and
attach/detach when specific listenables added or removed. The framework
helps you so:

=over 4

=item 1

The default implementations of C<attach()/detach()> makes or breaks all
connections to the listenables that you specify with the template method
explained below. C<attach()> is called by the framework when you create
your object, C<detach()> when you destroy it. It is up to you to call
them when the listenables are replaced, to give the framework a chance
to attach()/detach. If only one of the listenables is set or removed,
you can call C<attachTo()/detachFrom()> with one parameter: the name of
listenable.

If you want to use the default C<attach()/detach()> methods, you need to
provide one template method: C<getAttachments()>. The results will be
used to determine the attachments of this listener to listenable
objects. You must return a hash ref of listenable names. For each name
you provide a hash ref of event names that interest you on that
listenable, and for each event name an event handler name on your
object. 

Here is an example of a car listening to its parts: the door and the sun
roof:

  sub getAttachments : Protected
  {
     my $self = shift;
     return {
        door            => {
           Open         => 'StateChange',
           Close        => 'StateChange',
        }
        $self->hasSunRoof? (
           sunRoof      => {
              Open      => 'Open',
              Close     => 'Close',
           }
        ): ()
     };
  }

The car listens to the sun roof only if it has one. The car
listens to the Open and Close events of both parts. These will call the
following methods of the car: C<handleEvent_door_StateChanged()>,
C<handleEvent_sunRoof_Open()>, and C<handleEvent_sunRoof_Close()>.

The framework will attach/detach automatically when your listener object
is created and destroyed. However you must call
C<attachTo()/detachFrom()> when a door is being replaced with a new one,
or when changing the sun roof. 

You must provide a getter for each listenable name in
C<getAttachments()>, so that the framework can find the listenable by
name. Thus in the above example you would need to provide the methods
C<door()>, and C<sunRoof()>, that return the named listenables.

=item 2

You can override the default attachment methods if they are not right
for you. L<Rui::Model::List>, for example, may need to attach to all its
children. This is difficult to specify using C<getAttachments()>, so it
overrides C<attach()/detach()>, and implements its own mechanism,
without using C<getAttachments()>.

=back

=head1 METHODS

=head2 CONSTRUCTING

=head3 new

=over 4

=item parameters

Hash of named parameters. Keys/values defined here:

events - Optional. Hash ref. keys are event names, values are listeners. 

=back

=head2 LISTENING

=head3 addListener

=over 4

=item parameters

eventName - Scalar. What is the name of the event that interests me?

listener - callback, object, or method/object pair.

=item returns

Self.

=item description

Add a listener to the event name specified. The listener will be called with
event as only parameter. 

=back

=head3 removeListener

=over 4

=item parameters

eventName - Scalar. What is the name of the event that interests me?

listener - callback, object, or method/object pair.

=item returns

Self.

=item description

Remove a listener for the named event. Throws exception if listener not
found.

=back
