
package Rui::Event::Tree;

use base qw(Rui::Event);

sub init : Protected
{
	my ($self, %params) = @_;
	$self->SUPER::init(%params);
	$self->{path}  = $params{path};
	$self->{size}  = $params{size};
	$self->{items} = $params{items};
	return $self;
}

sub path
{ 
	my $self = shift;
	return $self->{path} unless @_;
	$self->{path} = shift;
	return $self;
}

sub size
{ 
	my $self = shift;
	return $self->{size} unless @_;
	$self->{size} = shift;
	return $self;
}

sub items
{ 
	my $self = shift;
	return $self->{items} unless @_;
	$self->{items} = shift;
	return $self;
}

1;

=head1 NAME

Rui::Event::Tree - a Tree model change event

=head1 SYNOPSIS

  # 5 items were added at path 0,1
  my $event = Rui::Event::Tree->new(
     name     => 'RangeAdded',
     source   => $self,
     path     => [0,1],
     size     => 5,
     items    => [@items],
  );
  print join(',', @{$event->path}, $event->size); # 0,1,5
  $event->path([1,2,3]);
  $event->size(2);

=head1 SUPERCLASS

L<Rui::Event>

=head1 DESCRIPTION

A change in a tree. The event is described by an path in the tree, and
the number of tree items involved, starting from the last point in
the path.

A path is a list of indices that start from the root tree item. So the
path C<0,1,2> is pointing at: the 3rd child of the 2nd child of the 1st
item that is a child of the root tree item. The path is always expected
and returned as an array ref.

The root node is indicated by an empty path: C<[]>.

Event name will be one of:

=over 4

=item *

C<RangeAdded>

=item *

C<RangeRemoved>

=item *

C<RangeChanged> - entire tree nodes (value + children ) have been changed

=item *

C<RangeValueChanged> - the value of some tree nodes has changed

=back

=head1 METHODS

=head2 CONSTRUCTING

=head3 new

=over 4

=item parameters

Hash of named parameters. Keys used here:

path - Array ref of integers. Where did the change happen.

size - Integer>0. Number of elements involved.

items - Array ref. Items removed/added/changed.

=back

=head2 ACCESSING

=head3 path

Get/set path. Array ref of integers.

=head3 size

Get/set size. Integer.

=head3 items

Get/set items. Array ref. Items removed/added/changed.

