
package Rui::Model::Entity;

use Carp;
use Attribute::Protected;
use Rui::Event;
use Rui::Model::Value;
use Rui::Model::List;

use base 'Rui::Model::Base';

# class info ---------------------------------------------------------------

my $Properties; # set lazily

sub getSchema
{{
	abstract => 1,
}}

# creating -----------------------------------------------------------------

# only permissable way to create new entity
sub create
{
	my ($thing, %params) = @_;
	my $class = ref($thing) || $thing;
	my $self  = $class->new(%params);
	my %properties = %{$self->getProperties};
	foreach my $name (keys %properties) {
		my $property = $properties{$name};
		$self->{$name} =
			exists $params{$name}       ? $params{$name}:
			exists $property->{default} ? $property->{default}:
			undef;
	}
	return $self;
}

sub erase
{
	my $self = shift;
	$self->SUPER::erase;
	foreach my $holderName (keys %{$self->getProperties}) {
		my $holder = delete $self->{"holder_$holderName"};
		if ($holder) {
			my $propertySpec = $self->getProperties->{$holderName};
			my $type         = $propertySpec->{type};
			my $aggreg       = $propertySpec->{aggreg};
			if (($type eq 'arrray' || $type eq 'iarray') && $aggreg) {
				foreach my $child ($holder->getAll) { $child->erase }
			}
			$holder->erase;
		}
		delete $self->{$holderName};
	}
}

# accessing ---------------------------------------------------------------

# wrap entity with value model
sub asValue { Rui::Model::Value->new(value => shift) }

# property handling --------------------------------------------------------

# only permissable way for outside world to change property
sub getHolder
{
	my ($self, $name) = @_;
	$self->checkPropertyName($name);
	my $holderName = "holder_$name";
	return $self->{$holderName} if exists $self->{$holderName};
	my $propertySpec = $self->getProperties->{$name};
	croak "unknown property '$name'" unless $propertySpec;

	my $holderClass = $propertySpec->{holderClass};
	croak "wrong field schema '$name'" unless $holderClass;
	$self->{$holderName} = $holderClass->new(
		adapt     => [$name => $self],
		weakAdapt => 1,
		attach    => 1, # only relevant for Rui::Model::List
	);

	return $self->{$holderName};
}

# not really public, to be called only by property holders
sub property
{
	my ($self, $name, $value) = @_;
	$self->checkPropertyName($name);
	return $self->{$name} if @_ == 2;
	$self->{$name} = $value;
	$self->fireEvent(Rui::Event::Value->new(
		name   => 'Change',
		value  => $value,
		source => $self,
	));
	return $self;
}

# returns hash ref of properyName => propertyData
sub getProperties
{
	my $self = shift;
	return $Properties if defined $Properties;
	my %out                = ();
	my %fields             = $self->getISAChainSchemaFields;
	my %holderTypeMapping  = $self->getHolderTypeMapping;
	my %holderClassMapping = ();
	while (my ($class, $types) = each %holderTypeMapping) {
		$holderClassMapping{$_} = $class foreach @$types;
	}
	foreach my $type (keys %fields) {
		my $holderClass  = $holderClassMapping{$type};
		my %fieldsSchema = %{$fields{$type}};
		foreach my $name (keys %fieldsSchema) {
			$out{$name} = {
				holderClass => $holderClass,
				default     => $fieldsSchema{$name}->{default},
				type        => $type,
				aggreg      => $fieldsSchema{$name}->{aggreg} || 0,
			};
		}
	}
	return \%out;
}

sub getISAChainSchemaFields : Private
{
	my $self  = shift;
	my $class = ref($self);
	my %out   = ();
	no strict;
	my @isaChain = ($class);
	while (my $baseClass = ${"$class\::ISA"}[0]) {
		push @isaChain, $baseClass;
		last if $baseClass eq __PACKAGE__;
		$class = $baseClass;
	}
	foreach my $baseClass (@isaChain) {
		my $classSchema = $baseClass->getSchema;
		next unless exists $classSchema->{fields};
		my %fields = %{$classSchema->{fields}};
		foreach my $type (keys %fields) {
			$out{$type} = {} unless exists $out{$type};
			my $fieldsSchema = $fields{$type};
			$out{$type}->{$_} = $fieldsSchema->{$_}
				foreach keys %$fieldsSchema;
		}
	}
	return %out;
}

sub getHolderTypeMapping : Private
{(
	Rui::Model::Value => [qw(string int real ref)],
	Rui::Model::List  => [qw(iarray array)],
)}

sub checkPropertyName
{
	my ($self, $name) = @_;
	croak "unknown property '$name'"
		unless exists $self->getProperties->{$name};
}

sub AUTOLOAD
{
	my $self = shift;
	return if $AUTOLOAD =~ /DESTROY$/;
	$AUTOLOAD =~ s/^.*:://;
	if ($AUTOLOAD =~ s/^getChildModel_//) {
		$self->checkPropertyName($AUTOLOAD);
		return $self->getHolder($AUTOLOAD);
	}
	$self->checkPropertyName($AUTOLOAD);
	return $self->property($AUTOLOAD, @_);
}

# --------------------------------------------------------------------------

=head1 NAME

Rui::Model::Entity - base class for domain models

=head1 DESCRIPTION

Domain models are models that with the following features:

=over4

=item *

They have properties. These are defined per class in the C<getSchema()>
template method. Properties have a name, a tangram compatible type, and
a name of a model class of the holder. The holder is an adapter on the
property. The only interface for get/setting properties is by getting a
holder on the property (C<getHolder($propertyName)>), then working vs.
that holder. Through this holder you can also listen to events on the
property.

=item *

They have one event, Change, that fires when the object has changed. In
that respect they are like value models.

=item *

You create them with a call to C<create()>, not C<new()>.

=back

=cut

1;

