
package Rui::Model::List;

use Carp;
use Attribute::Protected;
use Rui::Event::List;

use base 'Rui::Model::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	my $data = exists $params{data}? $params{data}: [];
	$self->{attach} = $params{attach} || 0;
	$self->{dataHolder} = Rui::Model::Value->
		new(exists $params{adapt}? (adapt => $params{adapt}): ());
	$self->data($data)
		if exists $params{data} || !(exists $params{adapt});
	$self->SUPER::init(%params);
}

# accessing ----------------------------------------------------------------

sub getSize { scalar @{shift->data} }
sub getAll  {        @{shift->data} }

sub get
{
	my ($self, $index) = @_;
	# die unless index numeric
	croak 'No index defined' unless defined $index;
	croak "Index [$index] out of range"
		if $index < 0 || $index >= $self->getSize;
	return $self->data->[$index];
}

sub add
{
	my ($self, $items, $index) = @_;
	my $size = $self->getSize;
	$index = $size if @_ == 2;
	croak "Index [$index $size] out of range"
		if $index < 0 || $index > $size;

	my @items = ref $items eq 'ARRAY'? @$items: ($items);
	splice(@{$self->data}, $index, 0, @items);
	$self->attachToItem($_) foreach @items;
	$self->fireListEvent(RangeAdded => $index, scalar @items, \@items);
	$self->data(scalar $self->data); # tell input value has changed
	return $self;
}

sub remove
{
	my ($self, $index, $length) = @_;
	my $size = $self->getSize;
	croak "Cannot remove on empty list" if $size == 0;
	$length = 1 if @_ == 2;
	croak "Index [$index, $length] out of range"
		if $index < 0 || $index >= $size;
	croak "Length [$index, $length] out of range"
		if ($size - $index) < $length;

	my @items = splice(@{$self->data}, $index, $length);
	$self->detachFromItem($_) foreach @items;
	$self->fireListEvent(RangeRemoved => $index, $length, \@items);
	$self->data(scalar $self->data); # tell input value has changed
	return $self;
}

sub set
{
	my ($self, $item, $index) = @_;
	my $size = $self->getSize;
	croak"Cannot set on empty list" if $size == 0;
	$index = $size - 1 if @_ == 2;
	croak "Index [$index] out of range" if $index < 0 || $index >= $size;
	$self->remove($index, 1);
	$self->add($item, $index);
	return $self;
}

sub removeItem
{
	my ($self, $item) = @_;
	$self->remove($self->getIndexByItem($item));
}

sub clear
{
	my $self = shift;
	return unless $self->getSize;
	$self->remove(0, $self->getSize);
	return $self;
}

sub getIndexByItem
{
	my ($self, $item) = @_;
	croak 'Item not defined' unless defined $item;
	croak 'List empty' unless $self->getSize;
	my $index = 0;
	my @data = $self->data;
	$index++ until $index == @data || $item eq $data[$index];
	croak "Item [$item] not in list" unless $index < @data;
	return $index;
}

sub itemExists
{
	my ($self, $item) = @_;
	croak 'Item not defined' unless defined $item;
	return 0 unless $self->getSize;
	my $index = 0;
	my @data = $self->data;
	# need more flexible way to define equality
	$index++ until $index == @data || $item eq $data[$index];
	return $index < @data;
}

# event handling -----------------------------------------------------------

sub handleEvent_Change
{
	my ($self, $event) = @_;
	my $index = $self->getIndexByItem($event->source);
	$self->fireListEvent(RangeChanged => $index, 1, [$event->source]);
}

sub handleEvent_Erase
{
	my ($self, $event) = @_;
	$self->removeItem($event->source);
}

sub fireListEvent : Protected
{
	my ($self, $name, $index, $size, $items) = @_;
	$self->fireEvent(Rui::Event::List->new(
		name   => $name, index  => $index,
		size   => $size, source => $self,
		items  => $items,
	));
}

# attach-detach ------------------------------------------------------------

sub attach : Protected
{
	my $self = shift;
	$self->SUPER::attach;
	return unless $self->{attach};
	$self->attachToItem($_) foreach $self->data;
}

sub detach : Protected
{
	my $self = shift;
	$self->SUPER::detach;
	return unless $self->{attach};
	$self->detachFromItem($_) foreach $self->data;
}

sub attachToItem : Private
{
	my ($self, $item) = @_;
	croak "no item" unless $item;
	return unless $self->{attach};
	$item->addListener($_ => $self) foreach $self->getItemEventNames;
}

sub detachFromItem : Protected
{
	my ($self, $item) = @_;
	return unless $self->{attach};
	$item->removeListener($_ => $self) foreach $self->getItemEventNames;
}

sub getItemEventNames : Protected { qw(Change Erase) }

# adapting -----------------------------------------------------------------

sub data : Private
{
	my $self = shift;
	unless (@_) {
		my $value = $self->{dataHolder}->value;
		croak "value [$value] not array ref" unless ref($value) eq 'ARRAY';
		return wantarray? @$value: $value;
	}
	$self->{dataHolder}->value(shift);
	return $self;
}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Model::List - a listenable collection

=head1 SYNOPSIS

  $list = Rui::Model::List->new;

  $list = Rui::Model::List->new(
     data    => qw[ a b ],
     attach  => 0, # not needed because 0 is default
     events  => {
        RangeAdded   => $self,
        RangeRemoved => $self,
        RangeChanged => $self,
     },
  );
                                    # List State    Event Fired
                                    # ----------    -----------
  $list->set('foo', 1);             # a foo         RangeChanged [index=1, size=1]
  $list->add('x');                  # a foo x       RangeAdded   [index=1, size=1]
  $list->add(['y', 'z'], 0);        # y z a foo x   RangeAdded   [index=0, size=2]
  $list->remove(0, 3);              # foo x         RangeRemoved [index=0, size=3]
  $list->remove(0);                 # x             RangeRemoved [index=0, size=1]

  print $list->getIndexByItem('x'); # 0
  print $list->get(0);              # x
  print $list->itemExists('foo');   # 0
  print $list->getSize;             # 1

  # example of list attaching to item events and re-broadcasting them
  $list = Rui::Model::List->new(
     data    => [map {Rui::Model::Value->new} (0..2)],
     attach  => 1, 
  );
  $list->addListener(RangeChanged => $self);
  $list->get(0)->data('foo');      #               RangeChanged [index=0, size=1]

=head1 SUPERCLASS

L<Rui::Model::Base>

=head1 REQUIRES

L<Rui::Event::List>

=head1 EVENTS

=over 4

=item RangeAdded

=item RangeRemoved

=item RangeChanged

L<Rui::Event::List>. The list has been modified. C<< $event->index >> is
the first index of change in the list, C<< $event->data >> is the
number of relevant elements.

=back

=head1 DESCRIPTION

B<Note>: you cannot put undef values in the list. Use a
L<value model|Rui::Model::Value> with an undef value.

If C<attach> is true, then you can only put objects that feature
C<Change> and C<Erase> events in the list (something like a L<value
model|Rui::Model::Value>. The list will listen to these events in its
children, and update the list listeners.

If C<attach> is false, you can put anything in the list, except undef
values.

C<RangeChanged> will only be fired when C<attach> is true.
