
package Rui::Model::SelectionInList;

use Carp;
use Rui::Model::Value;
use Rui::Model::List;
use Attribute::Protected;

use base 'Rui::Model::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	$self->{isNavigating} = exists $params{isNavigating}?
		$params{isNavigating}: 1;
	$self->{listHolder} = $params{listHolder} || Rui::Model::Value->new(
		value => $params{list} || Rui::Model::List->new(
			data   => $params{data}   || [],
			attach => $params{attach} || 0,
	));
	$self->{selectionHolder} = Rui::Model::Value->new(
		value => $params{selection} || Rui::Model::Value->new(
			value => exists $params{selectedItem}?
				$params{selectedItem}: exists $params{selectedIndex}?
					$self->{listHolder}->value->get($params{selectedIndex}):
					undef
	));
	$self->selectFirst if $params{selectFirst};
	$self->SUPER::init(%params);
}

# accessing ----------------------------------------------------------------

sub list
{
	my ($self, $list) = @_;
	return $self->{listHolder}->value if @_ == 1;
	$self->{listHolder}->value($list);
	return $self;
}

sub selection
{
	my ($self, $selection) = @_;
	return $self->{selectionHolder}->value if @_ == 1;
	$self->{selectionHolder}->value($selection);
	return $self;
}

sub listHolder
{
	my ($self, $listHolder) = @_;
	return $self->{listHolder} if @_ == 1;
	$self->detachFrom('listHolder');
	$self->{listHolder} = $listHolder;
	$self->attachTo('listHolder');
	$self->selectedItem(undef);
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'List_Change',
		value  => $self->list,
		source => $self,
	));
	$self->selectFirst if $self->{isNavigating};
	return $self;
}

sub selectionHolder { shift->{selectionHolder} }

# list management ----------------------------------------------------------

sub addNearSelection
{
	my ($self, $item, $isAfter) = @_;
	my $selectedItem = $self->selectedItem;
	my $list = $self->list;
	$list->add(
		$item,
		defined $selectedItem?
			$list->getIndexByItem($selectedItem) + $isAfter: 0
	);
	$self->selectedItem($item);
	return $self;
}

sub addBeforeSelection { shift()->addNearSelection(shift, 0) }
sub addAfterSelection  { shift()->addNearSelection(shift, 1) }

sub moveUp
{
	my $self  = shift;
	my $item  = $self->selectedItem;
	return unless $item;
	my $index = $self->selectedIndex;
	return unless $index;
	$self->list->remove($index);
	$self->list->add($item, $index - 1);
	$self->selectedIndex($index - 1);
	return $self;
}

sub moveDown
{
	my $self  = shift;
	my $item  = $self->selectedItem;
	return unless $item;
	my $index = $self->selectedIndex;
	return unless $index < ($self->list->getSize - 1);
	$self->list->remove($index);
	$self->list->add($item, $index + 1);
	$self->selectedIndex($index + 1);
	return $self;
}

sub removeSelection
{
	my $self = shift;
	$self->list->remove($self->selectedIndex);
	return $self;
}

# selection management -----------------------------------------------------

sub selectedIndex
{
	my ($self, $selectedIndex) = @_;
	my $selection = $self->selection;
	my $value = $selection->value;
	if (@_ == 1) {
		croak 'No item selected' unless defined $value;
		return $self->list->getIndexByItem($value);
	}
	$selection->value($self->list->get($selectedIndex));
  	return $self;
}

sub selectedItem
{
	my ($self, $newValue) = @_;
	my $selection = $self->selection;
	croak "no selection model" unless $selection;
	my $oldValue = $selection->value;
	return $oldValue if @_ == 1;

	return $self if 
		(!defined $newValue && !defined $oldValue) || (
			defined $newValue && defined $oldValue &&
			$oldValue eq $newValue
		);

	$selection->value($newValue);
	return $self;
}

sub selectFirst
{
	my $self = shift;
	return unless $self->list;
	$self->selectedIndex(0) if $self->list->getSize;
	return $self;
}

# event handling -----------------------------------------------------------

sub handleEvent_listHolder_Change
{
   my $self = shift;
   $self->selectedItem(undef);
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'List_Change',
		value  => $self->list,
		source => $self,
	));
	$self->selectFirst if $self->{isNavigating};
}

sub handleEvent_listHolder_RangeAdded
{
	my ($self, $event) = @_;
	$event->name('List_RangeAdded');
	$event->source($self);
	$self->fireEvent($event);
}

sub handleEvent_listHolder_RangeRemoved
{

	my ($self, $event) = @_;
	$event->name('List_RangeRemoved');
	$event->source($self);
	$self->fireEvent($event);
	return unless defined $self->selectedItem;
	return $self->selectedItem(undef) unless $self->list->getSize;
	return if $self->list->itemExists($self->selectedItem);

	# if we are still here then selected item has been removed from list
	return $self->selectedItem(undef) unless $self->{isNavigating};
	$self->selectedIndex($event->index == 0? 0: $event->index - 1);
}

sub handleEvent_listHolder_RangeChanged
{
	my ($self, $event) = @_;
	$event->name('List_RangeChanged');
	$event->source($self);
	$self->fireEvent($event);
}

sub handleEvent_selectionHolder_Change
{
   my $self = shift;
	$self->selection(Rui::Model::Value->new) unless $self->selection;
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'Selection_Change',
		value  => $self->selectedItem,
		source => $self,
	));
}

# template methods ---------------------------------------------------------

sub getAttachments : Protected
{{
	listHolder      => {
		Change             => 'Change',
		Value_RangeAdded   => 'RangeAdded',
		Value_RangeRemoved => 'RangeRemoved',
		Value_RangeChanged => 'RangeChanged',
	},
	selectionHolder => {
		Change             => 'Change',
		Value_Change       => 'Change',
	},
}}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Model::SelectionInList - selection in a list

=head1 SYNOPSIS

  $selectionInList = Rui::Model::SelectionInList->new(
     list            => Rui::Model::List->new(data => qw[ a b c ]),
     selection       => Rui::Model::Value->new(value => 'a'),
  );
  $selectionInList = Rui::Model::SelectionInList->new(
     data            => [qw( a b c )],
     selectedItem    => 'a',
     attach          => 0, # not needed because 0 is default
  );
  $selectionInList = Rui::Model::SelectionInList->new(
     data            => [qw( a b c )],
     selectedIndex   => 0,
     events          => {
        List_RangeAdded   => $listener,
        List_RangeRemoved => $listener,
        List_RangeChanged => $listener,
        List_Change       => $listener,
        Selection_Change  => $listener,
     },
  );
  $firstItemInList = $selectionInList->list->get(0);     # 'a'
  $selectedIndex = $selectionInList->selectedIndex;      # 0
  $selectionInList->selectedItem('b');                   # selection='b'
  $selectionInList->selection(Rui::Model::Value->new);   # selection undefined
  $selectionInList->selectedIndex(1);                    # selection='b'
  $selectionInList->list->remove(1);                     # selection='a'
  $self->addBeforeSelection('foo');                      # selection='foo'
  $self->addAfterSelection('bar');                       # selection='bar'
  $self->removeSelection;                                # selection='foo'
  $selectionInList->list(Rui::Model::List->new);         # selection undefined

=head1 SUPERCLASS

L<Rui::Model::Base>

=head1 REQUIRES

L<Rui::Model::Value>, L<Rui::Model::List>   

=head1 EVENTS

=over 4

=item List_Change

L<Rui::Event::Value>. The entire list has changed.

=item Selection_Change

L<Rui::Event::Value>. The selection has changed.

=item List_RangeAdded

=item List_RangeRemoved

=item List_RangeChanged

L<Rui::Event::List>. The list has been modified. C<< $event->index >> is
the first index of change in the list, C<< $event->size >> is the number
of relevant elements. The selection in list propagates all list events.

=back

=head1 DESCRIPTION

A model of a list and a selection from it. Provides a simple interface
for  managing the list and the selection.

The selection is automatically changd when it has become invalid because
of certain changes to the list. These list events can cause me to change
the selection:

=over 4

=item *

C<List_Change> will undefine the selection, as it is meaningless for the
new list.

=item *

C<List_RangeRemoved> if the selected item is removed, we will try to
select the previous item from the list. If there is no such item, we
select the first item in the list. If there is no such item, we undefine
the selection. This is for the default case, when C<isNavigating> is
true. When it is false, we simply undefine the selection.

=item *

C<List_RangeChanged> will undefine the selection if the selected item is
no longer in the list.

=back

The selection is advanced to the new item when calling
C<addBefore/AfterSelection()>.
