
package Rui::Model::SelectionInTree;

use Carp;
use Rui::Model::Value;
use Rui::Model::Tree;
use Attribute::Protected;

use base 'Rui::Model::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	$self->{isNavigating} = exists $params{isNavigating}?
		$params{isNavigating}: 1;
	$self->{treeHolder} = $params{treeHolder} || Rui::Model::Value->new(
		value => $params{tree} || Rui::Model::Tree->new(
			root => $params{root}   || undef
	));
	$self->{selectionHolder} = Rui::Model::Value->new(
		value => $params{selection} || Rui::Model::Value->new(
			value => exists $params{selectedNode}?
				$params{selectedNode}: exists $params{selectedPath}?
					$self->{treeHolder}->value->nodeAt(@{$params{selectedPath}}):
					undef
	));
	$self->selectFirst if $params{selectFirst};
	$self->SUPER::init(%params);
}

# accessing ----------------------------------------------------------------

sub tree
{
	my ($self, $tree) = @_;
	return $self->{treeHolder}->value if @_ == 1;
	$self->{treeHolder}->value($tree);
	return $self;
}

sub selection
{
	my ($self, $selection) = @_;
	return $self->{selectionHolder}->value if @_ == 1;
	$self->{selectionHolder}->value($selection);
	return $self;
}

sub treeHolder
{
	my ($self, $treeHolder) = @_;
	return $self->{treeHolder} if @_ == 1;
	$self->detachFrom('treeHolder');
	$self->{treeHolder} = $treeHolder;
	$self->attachTo('treeHolder');
	$self->selectedNode(undef);
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'Tree_Change',
		value  => $self->tree,
		source => $self,
	));
	$self->selectFirst if $self->{isNavigating};
	return $self;
}

sub selectionHolder { shift->{selectionHolder} }

# Tree management ----------------------------------------------------------

sub addNearSelection
{
	my ($self, $node, $isAfter) = @_;
	my $selectedNode = $self->selectedNode;
	croak 'cannot add near selection with no selection' unless $selectedNode;
	my $tree = $self->tree;
	my @selectedPath = @{$self->selectedPath};
	my $index = pop @selectedPath;
	$tree->listAt(@selectedPath)->add($node, $index + $isAfter);
	$self->selectedNode($node);
	return $self;
}

sub addBeforeSelection { shift()->addNearSelection(shift, 0) }
sub addAfterSelection  { shift()->addNearSelection(shift, 1) }

sub removeSelection
{
	my $self = shift;
	my @selectedPath = @{$self->selectedPath};
	my $index = pop @selectedPath;
	$self->tree->listAt(@selectedPath)->remove($index);
	return $self;
}

# selection management -----------------------------------------------------

sub selectedPath
{
	my ($self, $selectedPath) = @_;
	my $selection = $self->selection;
	my $value = $selection->value;
	if (@_ == 1) {
		croak 'No node selected' unless defined $value;
		return $self->tree->getPathByNode($value);
	}
	$selection->value($self->tree->nodeAt(@$selectedPath));
  	return $self;
}

sub selectedNode
{
	my ($self, $newValue) = @_;
	my $selection = $self->selection;
	croak "no selection model" unless $selection;
	my $oldValue = $selection->value;
	return $oldValue if @_ == 1;

	return $self if 
		(!defined $newValue && !defined $oldValue) || (
			defined $newValue && defined $oldValue &&
			$oldValue eq $newValue
		);

	$selection->value($newValue);
	return $self;
}

sub selectFirst
{
	my $self = shift;
	return unless $self->tree;
	$self->selectedPath([]) if $self->tree->rootNode;
	return $self;
}

# event handling -----------------------------------------------------------

sub handleEvent_treeHolder_Change
{
   my $self = shift;
   $self->selectedNode(undef);
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'Tree_Change',
		value  => $self->tree,
		source => $self,
	));
	$self->selectFirst if $self->{isNavigating};
}

sub handleEvent_treeHolder_RangeAdded
{
	my ($self, $event) = @_;
	$event->name('Tree_RangeAdded');
	$event->source($self);
	$self->fireEvent($event);
}

sub handleEvent_treeHolder_RangeRemoved
{
	my ($self, $event) = @_;
	my @path = eval { @{$event->path} };
	my $index = pop @path;
	$event->name('Tree_RangeRemoved');
	$event->source($self);
	$self->fireEvent($event);
	return unless defined $self->selectedNode;
	return $self->selectedItem(undef) unless $self->listAt(@path)->getSize;
	return if $self->tree->nodeExists($self->selectedNode);

	# if we are still here then selected item has been removed from list
	return $self->selectedItem(undef) unless $self->{isNavigating};
	$self->selectedPath([@path, ($index == 0? 0: $index - 1)]);
}

sub handleEvent_treeHolder_RangeChanged
{
	my ($self, $event) = @_;
	$event->name('Tree_RangeChanged');
	$event->source($self);
	$self->fireEvent($event);
}

sub handleEvent_selectionHolder_Change
{
   my $self = shift;
	$self->selection(Rui::Model::Value->new) unless $self->selection;
	$self->fireEvent(Rui::Event::Value->new(
	   name   => 'Selection_Change',
		value  => $self->selectedNode,
		source => $self,
	));
}

# template methods ---------------------------------------------------------

sub getAttachments : Protected
{{
	treeHolder      => {
		Change             => 'Change',
		Value_RangeAdded   => 'RangeAdded',
		Value_RangeRemoved => 'RangeRemoved',
		Value_RangeChanged => 'RangeChanged',
	},
	selectionHolder => {
		Change             => 'Change',
		Value_Change       => 'Change',
	},
}}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Model::SelectionInTree - selection in a Tree

=head1 SUPERCLASS

L<Rui::Model::Base>

=head1 REQUIRES

L<Rui::Model::Value>, L<Rui::Model::Tree>   

=cut
