package Rui::Model::Tree;

use Carp;
use Attribute::Protected;
use Rui::Util;
use Rui::Event::Tree;

use base 'Rui::Model::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	my $root = $params{root} || Rui::Model::TreeNode->new;
	($self->{childrenRoutes}, $self->{root}) = ref $root eq 'ARRAY'?
		@$root: (list => $root);
	$self->{nodes} = {}; # hash of node => parent node
	$self->{lists} = {}; # hash of list => owner node
	$self->SUPER::init(%params);
}

# composition --------------------------------------------------------------

sub nodeAt
{
	my ($self, @path) =  @_;
	my $node = $self->{root};
	my $level = 0;
	foreach my $index (@path) {
		my $list = $self->getListByNode($node);
		$node = $list->get($index);
		$level++;
	}
	return $node;
}

sub listAt
{
	my ($self, @path) =  @_;
	my $node = $self->{root};
	my $level = 0;
	my $list = $self->getListByNode($node);
	foreach my $index (@path) {
		$node = $list->get($index);
		$list = $self->getListByNode($node);
		$level++;
	}
	return $list;
}

sub rootNode   { shift->nodeAt }
sub rootList   { shift->listAt }
sub nodeExists { shift->{nodes}->{scalar shift} }

sub getListByNode
{
	my ($self, $node) = @_;
	my $childrenRoutes = $self->{childrenRoutes};
	my @tokens         = split /\./, $childrenRoutes;
	my $list           = Rui::Util::jumpByRoute($node, \@tokens);
	return $list;
}

sub getNodeByList { shift->{lists}->{scalar shift} }

# can only by called after tree has attached to all nodes
sub getPathByNode
{
	my ($self, $node) = @_;
	my @path = ();
	my $rootNode = $self->rootNode;
	while ($node ne $rootNode) {
		my $parentNode = $self->{nodes}->{scalar $node};
		croak "no parent for node [$node]" unless $parentNode;
		unshift @path, $self->getListByNode($parentNode)->getIndexByItem($node);
		$node = $parentNode;
	}
	return @path;
}

sub traverse
{
	my $self = shift;
	my ($a, $b, $c, $d) = @_;
	my ($code, $fromNode, $parentNode) = ref $a eq 'CODE'?
		($a, $b, $c): (sub { $b->$a(shift, shift) }, $c || undef, $d || undef);
	$self->traverseCode($code, $fromNode, $parentNode);
	return $self;
}

sub traverseCode : Private
{
	my ($self, $code, $fromNode, $parentNode) = @_;
	$fromNode   ||= $self->rootNode;
	$parentNode ||= undef;
	$code->($fromNode, $parentNode);
	$self->traverseCode($code, $_, $fromNode)
		foreach $self->getListByNode($fromNode)->getAll;
	return $self;
}

# event handling -----------------------------------------------------------

sub handleEvent_RangeAdded
{
   my ($self, $event) = @_;
	my $node  = $self->getNodeByList($event->source);
	my @items = @{$event->items};
	$self->attachToNodeRecursive($_, $node) foreach @items;
   $self->fireEvent(Rui::Event::Tree->new(
   	name   => 'RangeAdded',
   	path   => [$self->getPathByNode($node), $event->index],
   	size   => $event->size,
   	source => $self,
 	  items  => \@items,
   ));
}

sub handleEvent_RangeRemoved
{
   my ($self, $event) = @_;
	my $node = $self->getNodeByList($event->source);
	my $path = [$self->getPathByNode($node), $event->index];
	my @items = @{$event->items};
	$self->detachFromNodeRecursive($_, $node) foreach @items;
	$self->fireEvent(Rui::Event::Tree->new(
		name   => 'RangeRemoved',
		path   => $path,
		size   => $event->size,
		source => $self,
		items  => \@items,
	));
}

sub handleEvent_Change
{
   my ($self, $event) = @_;
	my $node = $event->source;
	my $path = [$self->getPathByNode($node)];
	$self->fireEvent(Rui::Event::Tree->new(
		name   => 'RangeChanged',
		path   => $path,
		size   => 1,
		source => $self,
		items  => [$node],
	));
}

# attach-detach ------------------------------------------------------------

sub attach : Protected
{
	my $self = shift;
	$self->SUPER::attach;
	$self->attachToNodeRecursive($self->rootNode);
}

sub detach : Protected
{
	my $self = shift;
	$self->SUPER::detach;
	$self->detachFromNodeRecursive($self->rootNode);
}

sub attachToNodeRecursive : Private
{
	my ($self, $node, $parentNode) = @_;
	$self->traverse(attachToNode => $self, $node, $parentNode || undef);
}

sub detachFromNodeRecursive : Private
{
	my ($self, $node, $parentNode) = @_;
	$self->traverse(detachFromNode => $self, $node, $parentNode || undef);
}

sub attachToNode : Private
{
	my ($self, $node, $parentNode) = @_;
	my $list = $self->getListByNode($node);
	# this will not serialize
	$self->{nodes}->{scalar $node} = $parentNode || 1;
	$self->{lists}->{scalar $list} = $node;

	$node->addListener($_ => $self) foreach $self->getNodeEventNames;
	$list->addListener($_ => $self) foreach $self->getListEventNames;
}

sub detachFromNode : Private
{
	my ($self, $node, $parentNode) = @_;
	my $list = $self->getListByNode($node);

	delete $self->{nodes}->{scalar $node};
	delete $self->{lists}->{scalar $list};

	$node->removeListener($_ => $self) foreach $self->getNodeEventNames;
	$list->removeListener($_ => $self) foreach $self->getListEventNames;
}

sub getNodeEventNames : Protected { qw(Change) }
sub getListEventNames : Protected { qw(RangeAdded RangeRemoved) }

# --------------------------------------------------------------------------

1;

