package Rui::Model::Value;

use Carp;
use Rui::Util;
use Rui::Event::Value;
use Attribute::Protected;
use WeakRef;

use base 'Rui::Model::Base';

# creating -----------------------------------------------------------------

sub init
{
	my ($self, %params) = @_;
	$self->{adapt} = exists $params{adapt}? $params{adapt}: undef;
	weaken($self->{adapt}->[1]) if $params{weakAdapt};
	$self->{attachToAdaptee} = 1 if $params{attachToAdaptee};
	$self->adapteeValue($params{value}) if exists $params{value};
	$self->SUPER::init(%params);
}

# accessing ----------------------------------------------------------------

sub value
{
	my ($self, $newValue) = @_;
	my $oldValue = $self->adapteeValue;
	return $oldValue if @_ == 1;

	# no need to do anything if my value has not changed
	# if it is a ref value, fire event anyway, as we cannot know
	# if it has changed
	# need better way to test for equality, what if 1.0 and 1?
	return $self if 
		(!defined $newValue && !defined $oldValue) || (
			defined $newValue && defined $oldValue &&
			!ref($oldValue) && !ref($newValue) &&
			$oldValue eq $newValue
		);

	$self->detachFrom('value');
	$self->adapteeValue($newValue);
	$self->attachTo('value');

	$self->fireEvent(Rui::Event::Value->new(
		name   => 'Change',
		value  => $newValue,
		source => $self,
	));
	return $self;
}

# listening ----------------------------------------------------------------

sub addListener
{
	my ($self, $eventName, $listener) = @_;
	my $value = $self->value;
	my $isValueEvent = ($eventName =~ /^Value_/) && defined $value;
	(my $cleanName = $eventName) =~ s/^Value_//;
	$value->addListener($cleanName, [handleEvent_value_propagate => $self]) if
		$isValueEvent && !$self->hasListenerForEvent($eventName) &&
		$self->isAttachOk($cleanName, value => $value);
	$self->SUPER::addListener($eventName, $listener);
}

sub removeListener
{
	my ($self, $eventName, $listener) = @_;
	my $value = $self->value;
	my $isValueEvent = ($eventName =~ /^Value_/) && defined $value;
	(my $cleanName = $eventName) =~ s/^Value_//;
	$value->removeListener($cleanName, $self) if
		$isValueEvent && !$self->hasListenerForEvent($eventName) &&
		$self->isAttachOk($cleanName, value => $value);
	$self->SUPER::removeListener($eventName, $listener);
}

# event handling -----------------------------------------------------------

sub handleEvent_value_Erase { shift->value(undef) }

sub handleEvent_value_propagate : Protected
{
	my ($self, $event) = @_;
	$event->name("Value_". $event->name);
	$event->source($self);
	$self->fireEvent($event);
}

sub handleEvent_adaptee_Change
{
	my ($self, $event) = @_;
	$self->detachFrom('value');
	$self->attachTo('value');
	$event->source($self);
	$self->fireEvent($event);
}

# listeners that are on the value, not the value model
sub getValueEventNamesOfListeners : Protected
{
	 map { s/^Value_//; $_ } grep (/^Value_/, shift->getEventNamesOfListeners);
}

# adapting -----------------------------------------------------------------

sub adaptee
{
	my $self = shift;
	my $adapt = $self->{adapt};
	return undef unless $adapt;
	my $adaptee = $adapt->[1];
	return $adaptee;
}

sub adapteeValue : Private
{
	my ($self, $newValue) = @_;
	if (!$self->{adapt}) {
		# I am my own adaptee
		return $self->{value} if @_ == 1;
		$self->{value} = $newValue;
		return $self;
	}
 
	my ($route, $adaptee) = @{$self->{adapt}};
	croak "no route for adapting" unless $route;
	croak "adaptee not good for adapting: [$adaptee]" unless ref $adaptee;
	my @tokens    = split /\./, $route;
	my $lastToken = @_ == 1? undef: pop @tokens;
	my $target = Rui::Util::jumpByRoute($adaptee, \@tokens);
	return $target if @_ == 1;
	
	$lastToken =~ s/(^\{)|(\}$)//g?
		$target->{$lastToken} = $newValue:
		$target->$lastToken($newValue);
	return $self;
}

# template methods ---------------------------------------------------------

sub getAttachments : Protected
{
	my $self = shift;
	return {

		value => {
			'Erase' => 'Erase',
			map { $_ => 'propagate' } $self->getValueEventNamesOfListeners,
		},

		adaptee => $self->{attachToAdaptee}? {
			Change => 'Change',
		}: {},

	};
}

sub isAttachOk : Protected
{
	my ($self, $eventName, $listenableName, $listenable) = @_;
	return 1 if $listenableName eq 'adaptee';
	my $isRef = ref $listenable;
	return
		$eventName eq 'Erase'?
			$isRef && UNIVERSAL::isa($listenable, 'Rui::Model::Base')?
				1: 0:
			$isRef && UNIVERSAL::isa($listenable, 'Rui::Event::Listenable')?
				1: 0;
}

# --------------------------------------------------------------------------

=head1 NAME

Rui::Model::Value - the value model

=head1 SYNOPSIS

  $value = Rui::Model::Value->new(value => 'foo');
  $value->addListener(Change => $self);
  $value->value('bar');                  # Change event will be fired
  $value->removeListener(Change => $self);
  
  $model = Rui::Model::Value->new(
     value  => $value,
     events => {
        Change       => $self,
        Value_Change => $self,
     },
  );
  $value->value('baz');                  # Value_Change event will be fired
  $model->value(Rui::Model::Value->new); # Change event will be fired

  $adaptee = Car->new(doorCount => 5);
  $adapter = Rui::Model::Value->new(
     adapt => [doorCount => $adaptee],
     value => 3,                         # Car::doorCount($adaptee, 3) called
  );
  print $adapter->value;                 # 3, Car::doorCount($adaptee) called
  $adapter->value(4);                    # Car::doorCount($adaptee, 4) called

=head1 SUPERCLASS

L<Rui::Model::Base>

=head1 REQUIRES

L<Rui::Event::Value>

=head1 EVENTS

=over 4

=item Change

L<Rui::Event::Value>. Value has changed.

=item Value_EventName

You can listen to events on the value (not the value model) by attaching
C<Value_> in front of the name of the event. See the description.

=back

=head1 DESCRIPTION

A value model allows listeners to be notified when a single value changes.
When this happens, it will fire the C<Change> event.

The value model will also propagate any events from its value to its
listeners. If you add listeners on event names that start with
C<Value_>, when the value fires, the value model will also fire these
events. Example: if you want the value model to propagate C<DoorOpened>
events from its C<Car> value, listen to the event C<Value_DoorOpened> on
the value model.

Because you can listen to the value model and it will propagate events
from its value, there is no need to add/remove listeners when the value
is replaced. The value model does this for you. This simplifies the
value net. If you were listening to the value of the value model, you
would have to attach/detach when it is changed.

If you set the C<< adapt => [$route => $adaptee] >> key when
constructing, the value model will become an adapter on an aspect,
indicated by the router string, of the adaptee. Instead of holding its
value in itself, the model will get/set it from the adaptee using the
given route.

A value model will listen to the C<Erase> event of its value, if the
value is a model.

=head1 METHODS

=head2 CONSTRUCTING

=head3 new

=over 4

=item parameters

Hash of named parameters. Keys/values defined here:

B<value> - Optional. Any. My value.

B<adapt> - Array ref of 2 elements. 1st is a route on the adaptee, 2nd is
the adaptee. When getting/setting value, the value model will get/set
its value from the adaptee, using the given route. The last segment of
the route will be set with the new value when setting. Example:

  Rui::Model::Value->new(adapt => ['value.{key}.property' => $foo]);

Will create an adapter that gets/sets its value from the object C<$foo>
using the call chain: C<< $foo->value->{key}->property >>.

B<attachToAdaptee> - Boolean, default false. If true the value model
will propagate all C<Change> events from its adaptee, as if they where
events of the value model. This is usefull as it makes the value model
look more like the adaptee for its clients.

=back

=head2 ACCESSING

=head3 value

=over 4

=item description

Get/set value. Any.

=back

=cut

1;
