package Rui::Model::tests::List;

use base qw(Rui::Model::tests::ModelTestCase);

use Rui::Model::Value;
use Rui::Model::List;

sub getModelClass { 'Rui::Model::List' }

sub test_new
{
	my $self = shift;
	$self->makeModel(data => [qw(a b c)]);
	$self->assert_equals('a', $self->{model}->get(0));
	$self->assert_equals('b', $self->{model}->get(1));
	$self->assert_equals('c', $self->{model}->get(2));
	$self->assert_equals(3, $self->{model}->getSize);
}

sub test_set
{
	my $self = shift;
	my ($removeEvent, $addEvent);
	$self->makeModel(
		data   => [qw(a b c)],
		events => {
			RangeAdded   => sub { $addEvent    = shift },
			RangeRemoved => sub { $removeEvent = shift },
			RangeChanged => sub { ($removeEvent, $addEvent) = (0, 0) },
		},
	);
	$self->{model}->set('foo', 1);
	$self->assert_equals('foo', $self->{model}->get(1), 'get');
	$self->assert_equals(3, $self->{model}->getSize);
	$self->assert_equals(1, $removeEvent->index, 'removeEvent index');
	$self->assert_equals(1, $removeEvent->size, 'removeEvent size');
	$self->assert_equals(1, $addEvent->index, 'addEvent index');
	$self->assert_equals(1, $addEvent->size, 'addEvent size');
}

sub test_add
{
	my $self = shift;
	my $event = 0;
	$self->{model}->addListener(RangeAdded   => sub { $event = shift });
	$self->{model}->addListener(RangeRemoved => sub { $event = 0 });
	$self->{model}->addListener(RangeChanged => sub { $event = 0 });
	$self->{model}->add('c');
	$self->{model}->add([qw(a b)], 0);
	$self->assert($self->{model}->get(0) eq 'a');
	$self->assert($self->{model}->get(1) eq 'b');
	$self->assert($self->{model}->get(2) eq 'c');
	$self->assert($self->{model}->getSize == 3);
	$self->assert($event->index == 0, $event->index);
	$self->assert($event->size == 2, $event->size);
}

sub test_remove
{
	my $self = shift;
	my $event = 0;
	$self->makeModel(
		data   => [qw(a b c)],
		events => {RangeAdded => sub { $event = 0 }},
	);
	$self->{model}->addListener(RangeRemoved => sub { $event = shift });
	$self->{model}->addListener(RangeChanged => sub { $event = 0 });

	$self->{model}->remove(2);
	$self->assert_equals(2, $event->index);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('c', join '.', @{$event->items});
	$self->assert_equals(2, $self->{model}->getSize);

	$self->{model}->remove(0,2);
	$self->assert_equals(0, $event->index);
	$self->assert_equals(2, $event->size);
	$self->assert_equals('a.b', join '.', @{$event->items});
	$self->assert_equals(0, $self->{model}->getSize);
}

sub test_getIndexByItem
{
	my $self = shift;
	$self->{model}->add([qw(a b c)]);
	$self->assert($self->{model}->getIndexByItem('a') == 0);
	$self->assert($self->{model}->getIndexByItem('b') == 1);
	$self->assert($self->{model}->getIndexByItem('c') == 2);
}

sub test_itemExists
{
	my $self = shift;
	$self->{model}->add([qw(a b c)]);
	$self->assert($self->{model}->itemExists('a'));
	$self->assert($self->{model}->itemExists('b'));
	$self->assert($self->{model}->itemExists('c'));
	$self->assert(!$self->{model}->itemExists('d'));
}

sub test_attach_false
{
	my $self = shift;
	my $event = 0;
	$self->makeModel(
		data   => [
			Rui::Model::Value->new(value => 'a'),
			Rui::Model::Value->new(value => 'b'),
			Rui::Model::Value->new(value => 'c'),
		],
		attach => 0,
	);
	$self->{model}->addListener(RangeChanged => sub { $event = shift });
	$self->{model}->addListener(RangeAdded   => sub { $event = shift });
	$self->{model}->addListener(RangeRemoved => sub { $event = shift });
	$self->{model}->get(1)->value('foo');
	$self->assert( $event == 0 );
}

sub test_attach_true
{
	my $self = shift;
	my $event = 0;
	$self->makeModel(
		data   => [
			Rui::Model::Value->new(value => 'a'),
			Rui::Model::Value->new(value => 'b'),
			Rui::Model::Value->new(value => 'c'),
		],
		attach => 1,
	);
	$self->{model}->addListener(RangeChanged => sub { $event = shift });
	$self->{model}->addListener(RangeAdded   => sub { $event = 0 });
	$self->{model}->addListener(RangeRemoved => sub { $event = 0 });
	$self->{model}->get(1)->value('foo');
	$self->assert_equals(1, $event->index);
	$self->assert_equals(1, $event->size);
}

sub test_adapt
{
	my $self = shift;
	my $input = Rui::Model::Value->new(value => [qw(x y z)]);
	$self->makeModel(data => [qw(a b c)], adapt => [value => $input]);
	$self->assert_equals('a', $self->{model}->get(0));
	$self->{model}->remove(0);
	$self->assert_equals('b', $self->{model}->get(0));
	$self->assert_equals('b', $input->value->[0]);
	splice(@{$input->value}, 0, 1);
	$self->assert_equals('c', $self->{model}->get(0));
	$self->assert_equals('c', $input->value->[0]);
}

sub test_2Listeners
{
	my $self = shift;
	$self->makeModel(data => [qw(a b c)]);
	my ($event1, $event2) = (0, 0);
	$self->{model}->addListener(RangeAdded => sub { $event1 = shift });
	$self->{model}->addListener(RangeAdded => sub { $event2 = shift });
	$self->{model}->add(['foo']);
	$self->assert_equals(3, $event1->index, 'listener 1');
	$self->assert_equals(3, $event2->index, 'listener 2');
}

sub test_eraseChild
{
	my $self = shift;
	my $event = 0;
	my $value = Rui::Model::Value->new->value('b');
	$self->makeModel(
		data   => [
			Rui::Model::Value->new(value => 'a'),
			$value,
			Rui::Model::Value->new(value => 'c'),
		],
		attach => 1,
	);
	$self->{model}->addListener(RangeChanged => sub { $event = 0 });
	$self->{model}->addListener(RangeAdded   => sub { $event = 0 });
	$self->{model}->addListener(RangeRemoved => sub { $event = shift });
	$value->erase;
	$self->assert_equals(1, $event->index);
	$self->assert_equals(1, $event->size);
	$self->assert_equals(2, $self->{model}->getSize);
}

sub test_sharedValueEvents
{
	my $self = shift;
	my $list = Rui::Model::List->new;
	my ($event1, $event2) = (0, 0);
	my $model1 = Rui::Model::Value->new(value => $list);
	my $model2 = Rui::Model::Value->new(value => $list);
	$model1->addListener(Value_RangeAdded => sub { $event1 = shift });
	$model2->addListener(Value_RangeAdded => sub { $event2 = shift });
	$list->add('foo');
	$self->assert_equals(0, $event1->index, 'event1');
	$self->assert_equals(0, $event2->index, 'event2');
}

1;
