package Rui::Model::tests::Tree;

use Rui::Model::Value;
use Rui::Model::TreeNode;
use Rui::Model::Tree;

use base qw(Rui::Model::tests::ModelTestCase);

sub getModelClass { 'Rui::Model::Tree' }

sub makeBigModel
{
	my $self = shift;
	my $protoNode = Rui::Model::TreeNode->new;
	$self->makeModel(root => $protoNode->new(
		value => 'RootNode',  
		data  => [
			$protoNode->new(
				value => 'Child 1',
				data  => [
					$protoNode->new(value => 'Grandchild 1.1'),
					$protoNode->new(value => 'Grandchild 1.2'),
				],
			),
			$protoNode->new(
				value => 'Child 2',
				data  => [
					$protoNode->new(value => 'Grandchild 2.1'),
					$protoNode->new(value => 'Grandchild 2.2'),
					$protoNode->new(value => 'Grandchild 2.3'),
				],
			
			),
		],
	));
	return $self->{model};
}

sub test_newWithChildren
{
	my $self = shift;
	my $model = $self->makeBigModel;

	$self->assert_equals('RootNode',       $model->nodeAt->value);
	$self->assert_equals('Child 1',        $model->nodeAt(0)->value);
	$self->assert_equals('Child 2',        $model->nodeAt(1)->value);
	$self->assert_equals('Grandchild 1.1', $model->nodeAt(0, 0)->value);
	$self->assert_equals('Grandchild 2.3', $model->nodeAt(1, 2)->value);

	$self->assert_equals(2, $model->listAt->getSize);
	$self->assert_equals(2, $model->listAt(0)->getSize);
	$self->assert_equals(3, $model->listAt(1)->getSize);
	$self->assert_equals(0, $model->listAt(0, 0)->getSize);
	$self->assert_equals(0, $model->listAt(0, 1)->getSize);
	$self->assert_equals(0, $model->listAt(1, 0)->getSize);
	$self->assert_equals(0, $model->listAt(1, 1)->getSize);
	$self->assert_equals(0, $model->listAt(1, 2)->getSize);
}

sub test_traverse
{
	my $self = shift;
	my $model = $self->makeBigModel;

	$model->traverse(sub {
		my ($node, $parentNode) = @_;
		$node->value(lc $node->value);
	});

	$self->assert_equals('rootnode',       $model->nodeAt->value);
	$self->assert_equals('child 1',        $model->nodeAt(0)->value);
	$self->assert_equals('child 2',        $model->nodeAt(1)->value);
	$self->assert_equals('grandchild 1.1', $model->nodeAt(0, 0)->value);
	$self->assert_equals('grandchild 2.3', $model->nodeAt(1, 2)->value);
}

sub test_getPathByNode
{
	my $self = shift;
	my $model = $self->makeBigModel;
	my $pathByNode = sub {
		join '.', $model->getPathByNode( $model->nodeAt(@_) );
	};

	$self->assert_equals('',    $pathByNode->());
	$self->assert_equals('0',   $pathByNode->(0));
	$self->assert_equals('1',   $pathByNode->(1));
	$self->assert_equals('1.2', $pathByNode->(1, 2));
	$self->assert_equals('1.1', $pathByNode->(1, 1));
}

sub test_getPathByNodeOnAddedNodes
{
	my $self = shift;
	my $model = $self->makeBigModel;
	my $pathByNode = sub {
		join '.', $model->getPathByNode( $model->nodeAt(@_) );
	};

	$model->listAt(1, 1)->add( Rui::Model::TreeNode->new(value => 'new') );
	$self->assert_equals('1.1.0', $pathByNode->(1, 1, 0));
}

sub test_rangeAdded1
{
	my $self = shift;
	my $model = $self->makeBigModel;
	my $newNode = Rui::Model::TreeNode->new(value => 'new');

	my $event = 44;
	$model->addListener(RangeAdded => sub { $event = shift });

	$model->listAt->add($newNode);
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('2', join '.', @{$event->path});
}

sub test_rangeAdded2
{
	my $self = shift;
	my $model = $self->makeBigModel;
	my $newNode = Rui::Model::TreeNode->new(value => 'new');

	my $event = 44;
	$model->addListener(RangeAdded => sub { $event = shift });

	$model->listAt(1, 2)->add($newNode);
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1.2.0', join '.', @{$event->path});
}

sub test_rangeAddedOnAddedNodes1
{
	my $self = shift;
	my $model      = $self->makeBigModel;
	my $parentNode = Rui::Model::TreeNode->new(value => 'parent');
	my $childNode  = Rui::Model::TreeNode->new(value => 'child');

	my $event = 44;
	$model->addListener(RangeAdded => sub { $event = shift });
	$model->listAt(1, 2)->add($parentNode);
	$event = 44;
	$model->listAt(1, 2, 0)->add($childNode);

	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1.2.0.0', join '.', @{$event->path});
}

sub test_rangeAddedOnAddedNodes2
{
	my $self = shift;
	my $model      = $self->makeBigModel;
	my $parentNode = Rui::Model::TreeNode->new(value => 'parent');
	my $childNode  = Rui::Model::TreeNode->new(value => 'child');
	my $infantNode = Rui::Model::TreeNode->new(value => 'infant');

	my $event = 44;
	$model->addListener(RangeAdded => sub { $event = shift });
	$model->listAt(1, 2)->add($parentNode);
	$model->listAt(1, 2, 0)->add($childNode);
	$event = 44;
	$model->listAt(1, 2, 0, 0)->add($infantNode);

	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1.2.0.0.0', join '.', @{$event->path});
}

sub test_rangeRemovedOneNode
{
	my $self = shift;
	my $model = $self->makeBigModel;

	my $event = 44;
	$model->addListener(RangeRemoved => sub { $event = shift });

	$model->listAt(1)->remove(0);
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1.0', join '.', @{$event->path});
}

sub test_rangeRemovedTwoNodes
{
	my $self = shift;
	my $model = $self->makeBigModel;

	my $event = 44;
	$model->addListener(RangeRemoved => sub { $event = shift });

	$model->listAt(1)->remove(0, 2);
	$self->assert_equals($model, $event->source);
	$self->assert_equals(2, $event->size);
	$self->assert_equals('1.0', join '.', @{$event->path});
}

sub test_rangeChanged
{
	my $self = shift;
	my $model = $self->makeBigModel;

	my $event = 44;
	$model->addListener(RangeChanged => sub { $event = shift });

	$model->nodeAt(1)->value('new1');
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1', join '.', @{$event->path});

	$event = 44;
	$model->nodeAt(1, 2)->value('new1');
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('1.2', join '.', @{$event->path});

	$event = 44;
	$model->rootNode->value('new1');
	$self->assert_equals($model, $event->source);
	$self->assert_equals(1, $event->size);
	$self->assert_equals('', join '.', @{$event->path});

}

sub test_nodeExists
{
	my $self = shift;
	my $model = $self->makeBigModel;
	$self->assert($model->nodeExists($model->rootNode), 'root');
	$self->assert($model->nodeExists($model->nodeAt(1,2)), 'grandchild');
}

1;

__END__

