package Rui::Model::tests::Value;

use Data::Dumper;
use Rui::Model::Value;
use Rui::Model::List;

use base 'Rui::Model::tests::ModelTestCase';

sub getModelClass { 'Rui::Model::Value' }

sub test_value
{
	my $self = shift;
	$self->makeModel(value => 'foo');
	$self->assert_equals('foo', $self->{model}->value);
	$self->{model}->value('bar');
	$self->assert_equals('bar', $self->{model}->value);
}

sub test_events
{
	my $self = shift;
	my $callback_called = 0;
	my $callback = sub { $callback_called = shift()->value };
	$self->{model}->addListener(Change => $callback);
	$self->{model}->value('foo');
	$self->assert($callback_called eq 'foo');
	$self->{model}->removeListener(Change => $callback);
	$self->{model}->value('bar');
	$self->assert($callback_called eq 'foo');
}

sub test_newWithEvents
{
	my $self = shift;
	my $callback_called = 0;
	$self->makeModel(
		value  => 'foo',
		events => {Change => sub { $callback_called = shift()->value }},
	);
	$self->{model}->value('bar');
	$self->assert($callback_called eq 'bar');
}

sub test_valueEvents
{
	my $self = shift;
	my $event = 0;
	my $value = Rui::Model::Value->new(value => 'foo');
	$self->makeModel(
		value  => $value,
		events => {
			Change       => sub { $event = 1 },
			Value_Change => sub { $event = shift },
		},
	);
	$value->value('bar');
	$self->assert(
		$event->value eq 'bar' &&
		$event->name  eq 'Value_Change'
	);
	$self->{model}->value(undef);
}

sub test_stickyValueEvents
{
	my $self = shift;
	my $event = 0;
	my $value1 = Rui::Model::Value->new(value => 'foo');
	my $value2 = Rui::Model::Value->new(value => 'bar');
	$self->makeModel(
		value  => $value1,
		events => {Value_Change => sub { $event = shift }},
	);

	$self->{model}->value($value2);
	$value2->value('baz');
	$self->assert_equals('baz', $event->value, 'new model 1, value');
	$self->assert_equals('Value_Change', $event->name, 'new model 1, name');

	$event = 0;
	$value1->value('pakpak');
	$self->assert_equals(0, $event, 'detached 1');
	$self->{model}->value(undef);
	$value2->value('zifzif');
	$self->assert_equals(0, $event, 'detached 2');

	$self->{model}->value($value1);
	$value1->value('foofoo');
	$self->assert_equals('foofoo', $event->value, 'new model 2, value');
	$self->assert_equals('Value_Change', $event->name, 'new model 2, name');

	$self->{model}->value(undef);
}

sub test_adapt
{
	my $self = shift;
	my $adaptee = Data::Dumper->new([]);
	$self->makeModel(
		adapt => [Purity => $adaptee],
		value => 1,
	);
	$self->assert_equals(1, $adaptee->Purity, 'initial set');
	$self->assert_equals(1, $self->{model}->value, 'initial set from model');
	$self->{model}->value(0);
	$self->assert_equals(0, $adaptee->Purity, 'model set');
	$self->assert_equals(0, $self->{model}->value, 'model set from model');
	$adaptee->Purity(1);
	$self->assert_equals(1, $adaptee->Purity, 'adaptee set');
	$self->assert_equals(1, $self->{model}->value, 'adaptee set from model');
}

sub test_complexAdapt
{
	my $self = shift;
	my $stage1  = {stage1Key => 'foo'};
	my $stage2  = {stage2Key => $stage1};
	my $stage3  = Rui::Model::Value->new(value => $stage2);
	my $adaptee = Rui::Model::Value->new(value => $stage3);
	$self->makeModel(adapt => [
		'value.value.{stage2Key}.{stage1Key}' => $adaptee
	]);
	$self->assert_equals('foo', $self->{model}->value, 'initial set');
	$self->{model}->value('bar');
	$self->assert_equals('bar', $self->{model}->value, 'model value');
	$self->assert_equals('bar', $stage1->{stage1Key}, 'adaptee value');
}

sub test_attachToAdaptee
{
	my $self = shift;
	my $event = 0;
	my $adaptee = Rui::Model::Value->new(value => 'foo');
	$self->makeModel(
		adapt           => [value => $adaptee],
		attachToAdaptee => 1,
		events          => {Change => sub { $event = 1 }},
	);
	$adaptee->value('bar');
	$self->assert_equals(1, $event);
}

sub test_erase
{
	my $self = shift;
	my $value = Rui::Model::Value->new(value => 'foo');
	$self->makeModel(value => $value);
	$value->erase;
	$self->assert_null($self->{model}->value);
}

1;
