package Rui::Remote::Session;

use Carp;
use Attribute::Protected;
use Rui::Application::Map;

# creating -----------------------------------------------------------------

sub new
{
	my $thing = shift;
	my $self  = bless {}, ref($thing) || $thing;
	$self->init(@_);
	return $self;
}

sub init
{
	my ($self, $sessionId)   = @_;
	$self->{sessionId}       = $sessionId      || croak 'No sessionId';
	$self->{applicationMap}  = Rui::Application::Map->new;
	$self->{buffer}          = "";
	$self->{clientListeners} = {};
	$self->{application}     = undef;
	$self->resetIds;
}

sub destroy
{
	delete $self->{$_} foreach qw(application clientListeners);
}

sub resetIds
{
	my $self = shift;
	$self->{nextWidgetId}    = 0;
	$self->{nextWindowId}    = 0;
}

# server interface ---------------------------------------------------------

sub postClientEvent
{
	my ($self, %params) = @_;
	croak 'No requestType' unless $params{requestType}; 
	my $requestType = delete $params{requestType};
	if ($requestType eq 'boot') {
		$self->processBootRequest(%params);

	} elsif ($requestType eq 'event') {
		$self->processEventRequest(%params);

	} else {
		croak "Unknown requestType:[$requestType]";
	}
}

sub flushBuffer
{
	my $self = shift;
	my $buffer = $self->{buffer};
	$self->{buffer} = "";
	return "$self->{sessionId}\n$buffer";
}

# server widgets interface -------------------------------------------------

sub setClientListener    { shift->{clientListeners}->{shift()} = pop }
sub removeClientListener { delete shift->{clientListeners}->{shift()} }
sub setApplication       { shift->{application} = pop }
sub getApplication       { shift->{application} }
sub getId                { shift->{sessionId} }

sub addWidgetCommand
{
	my ($self, $widgetClass, $windowId, $widgetId, $methodName, @params) = @_;
	$self->addCommandLine($widgetClass, $methodName, $windowId, $widgetId, @params);
}

sub addWindowCommand
{
	my ($self, $windowClass, $windowId, $methodName, @params) = @_;
	$self->addCommandLine($windowClass, $methodName, $windowId, @params);
}

sub addCommandLine : Private
{
	my ($self, @params) = @_;
 	my $wordSeperator = chr(1);
 	my $lineSeperator = chr(2);
	$self->appendToBuffer(
		join($wordSeperator, map {
			croak "cannot send undefined params for: [". join(',',@params). "]"
				 unless defined $_;
			s/$lineSeperator/ /g;
			s/$wordSeperator/ /g;
			$_;
		} @params). $lineSeperator
	);
	return $self;
}

sub appendToBuffer : Protected { shift->{buffer} .= shift }

sub makeWindowId
{
	my $self = shift;
	$self->{nextWindowId}++;
	return "W$self->{nextWindowId}";
}

sub makeWidgetId
{
	my $self = shift;
	$self->{nextWidgetId}++;
	return "E$self->{nextWidgetId}";
}

# request handling ---------------------------------------------------------

sub processBootRequest : Private
{
	my ($self, %params) = @_;
	my $applicationName = $params{applicationName} || 'DEFAULT';
	delete $params{applicationName};
	my $applicationClassName = $self->{applicationMap}->
		getApplicationClassName($applicationName);
	if ($applicationClassName) {
		eval "use $applicationClassName;";
		croak "Error loading application class $applicationClassName:\n$@" if $@;

		# application will call setApplication on me
		$applicationClassName->new;
	} else {
		$self->sendError("No application class found for $applicationName");
	}
}

sub processEventRequest : Private
{
	my ($self, %params) = @_;
	my $sourceId = delete $params{sourceId} || croak 'Event with no sourceId';
	croak 'Event with no name' unless $params{name};
	my $listener = $self->{clientListeners}->{$sourceId};
	croak "Posthomous event from [$sourceId]" unless $listener;
	$listener->handleClientEvent(%params);
}

# utilities ----------------------------------------------------------------

# classic hack showing why we need proper client cought exceptions
sub sendError 
{
	my ($self, $error) = @_;
	$error ||= 'server could not process request';
	$self->appendToBuffer("\n>$error\n");
}

# --------------------------------------------------------------------------

=head1 NAME

Rui::Remote::Session - passive mediator between server and widgets

=head1 DESCRIPTION

There is a session for each association between a user and a running
application instance. The session is used by the server to post client
events to server widgets, and to get any messages they may wish to send
to the client. Server widgets use the session for passing messages to
the client, and listening to client events. It is a passive mediator
between them, and knows nothing about the server object or the server
widgets. The session lifecycle is handled by a session manager.

There are 2 message types that can be received from the client via the
server object: boot and event. Boot creates a new application, event
post an event from the client to some server widget. Posting a C<Close>
event for the main window destroys the session and the application.

Server side widgets register/deregister as listeners to client-side
events with C<setClientListener($id, $listener)> and
C<removeClientListener($id)>. The method
C<handleClientEvent(%eventParams)> will be called on the listener when a
client-side event is posted. The key used to connect server widgets with
client-side events is the window or widget id.

Messages to the client are send from server widgets to the client using
C<addWidget/WindowCommand>. The server flushes the messages using
C<flushBuffer()>, and sends them to the client.

Because there is exactly one session for each running application
instance, server widgets also use it to get/set their application
object, and to get the next available window and widget ids for the
session.

=cut

1;

