# This module should be used in socket-based servers and clients.
# Includes 2 message parsing methods for a client and 2 for the server.
# encodeRequest, decodeRequest, encodeResponce, decodeResponce
# 
# It provides a method reading from a socket readMessage;
# writting should be done using $socket->print.

package Rui::Remote::SocketMessages;
use strict;

my $EOLN = "\r\n";
my $EOLNre = "\r?\n";

# ----------------------------------------------------------------------------
# For use on client and server, this is the only method reading a socket.
# Reads the first message line that should be "Lines: N" and returns the rest.

sub readLine ($$) {
	my $this = shift;
	my $socket = shift;
	my $timeout = shift || 2 * 60 * 60;  # never wait more than 2 hours

	local $SIG{ALRM} = sub { die "timeout\n" };
	alarm $timeout if $^O ne 'MSWin32';
	my $line = eval { $socket->getline; };
	#print STDERR $@ if $@;
	alarm 0 if $^O ne 'MSWin32';
	return $line;
}

sub readMessage ($$;$) {
	my $this = shift;
	my $socket = shift;
	my $timeout = shift;

	my $linesString = $this->readLine($socket, $timeout);
	return undef unless defined $linesString;

	my $string = "";

	my $linesNum = $linesString =~ /^Lines: (\d+)/? $1: undef;
	# if the first line is not "Lines:" it is a foreign client, support it
	if (!defined $linesNum) {
		#die "Incorrect first message line: $linesString";
		$string = $linesString;
		$linesNum = 1e20;
	}

	foreach (my $i = 0; $i < $linesNum; $i++) {
		my $line = $this->readLine($socket, $timeout);
		last if !defined $line || $line =~ /^$EOLNre$/s && $linesNum == 1e20;
		$string .= $line;
	}
	return $string;
}

# ----------------------------------------------------------------------------
# for use on client only, converts hash to string

sub encodeRequest ($$) {
	my $this = shift;
	my $hash = shift;

	$this->setDefaultHeaders($hash);

	# evaluate number of lines in request message: only headers
	my $num_lines = keys(%$hash) + 1;

	my $string = "Lines: $num_lines$EOLN";
	foreach my $key (sort keys %$hash) {
		my $value = $hash->{$key};
		$string .= "$key: $value$EOLN";
	}
	$string .= "\n";

	return $string;
}

# ----------------------------------------------------------------------------
# for use on server only, converts string to hash

sub decodeRequest ($$) {
	my $this = shift;
	my $string = shift;
	return undef unless defined $string;

	$string =~ s/$EOLNre$//s;

	my $hash = {};
	foreach my $line (split(/$EOLNre/s, $string)) {
		my ($key, $value) = $line =~ /^(.*?):\s*(.*)$/;
		if (defined $key && defined $value) {
			$hash->{$key} = $value;
		} else {
			die "decodeRequest: Invalid header: $line\n";
		}
	}

	delete $hash->{'Lines'};
	$this->setDefaultHeaders($hash);

	return $hash;
}

# ----------------------------------------------------------------------------
# for use on server only, converts hash to string

sub encodeResponse ($$) {
	my $this = shift;
	my $hash = shift;

	$this->setDefaultHeaders($hash);

	my $content = $hash->{'Content'};
	$content = [] unless defined $content;
	delete $hash->{'Content'};
	if (ref($content) eq 'ARRAY') {
		push @$content, "" if @$content;
		$content = join($EOLN, @$content);
	}
	$content =~ s/$EOLNre/$EOLN/sg;

	# evaluate number of lines in response message: headers + content
	my $num_lines = keys(%$hash) + 1;
	$num_lines += length(join('', $content =~ /(\n)/sg)) + 1;

	my $string = "Lines: $num_lines$EOLN";
	foreach my $key (sort keys %$hash) {
		my $value = $hash->{$key};
		$string .= "$key: $value$EOLN";
	}
	$string .= $EOLN;
	$string .= $content;
	$string .= $EOLN;

	return $string;
}

# ----------------------------------------------------------------------------
# for use on client only, converts string to hash

sub decodeResponse ($$) {
	my $this = shift;
	my $string = shift;
	return undef unless defined $string;

	my ($headers, $content) = split(/$EOLNre$EOLNre/, $string, 2);
	$content = "" unless defined $content;
	$content =~ s/$EOLNre$//s;

	my $hash = {};
	foreach my $line (split(/$EOLNre/s, $headers)) {
		my ($key, $value) = $line =~ /^(.*?):\s*(.*)$/;
		if (defined $key && defined $value) {
			$hash->{$key} = $value;
		} else {
			die "decodeResponse: Invalid header: $line\n";
		}
	}
	$hash->{'Content'} = $content;

	delete $hash->{'Lines'};
	$this->setDefaultHeaders($hash);

	return $hash;
}

# ----------------------------------------------------------------------------

sub getUniqueIdString ($) {
	return time() . '-' . int(rand(1000000000));
}

sub getDateString ($) {
	return time() if $^O eq "MSWin32";
	return `date '+%Y-%m-%d %T' | tr -d '\n'`;
}

sub setDefaultHeaders ($$) {
	my $this = shift;
	my $hash = shift;

#	$hash->{'Id'} ||= $this->getUniqueIdString;
#	$hash->{'Date'} ||= $this->getDateString;
}

1;
