package Rui::Server::Base;
use strict;

use Carp;
use Attribute::Abstract;
use Attribute::Protected;
use Rui::Remote::SessionManager;
use Rui::Remote::ClientMessage;

sub new
{
	my ($this, %params) = @_;
	my $self  = bless {}, ref($this) || $this;
	$self->{logFile}        = delete $params{LogFile} || undef;
	$self->{pidFile}        = delete $params{PidFile};
	$self->{sessionManager} = Rui::Remote::SessionManager->instance;

	my $pidFile = $self->{pidFile};
	if ($pidFile) {
		open PID, ">$pidFile" || $self->log("Can't write to pid file $pidFile");
		print PID "$$\n";
		close PID;
	}

	return $self;
}

sub DESTROY {
	my $self = shift;
	my $pidFile = $self->{pidFile};
	unlink $pidFile if $pidFile;
}

sub run : Abstract;

sub handleApplicationEvent : Protected {
	my ($self, $event) = @_;

	my $sessionId = delete $event->{sessionId};
	my $session = $sessionId?
		$self->{sessionManager}->getSessionById($sessionId):
		$self->{sessionManager}->createSession;

	# hack: should report session expired to user, and not kill server
	return '' unless $session;

	$self->{sessionManager}->setActiveSession($session);
	$session->postClientEvent(%$event);
	my $content = $session->flushBuffer;
	$self->{sessionManager}->setActiveSession(undef);

	return $content;
}

sub log : Protected {
	my $self = shift;
	my $fileName = $self->{logFile};
	return if !$fileName;
	my $msg = join("\n", @_) . "\n";
	my $premsg = "";

	my $time;
	if (scalar(localtime) =~ /(\w+) ( ?\d+) (\d+:\d+:\d+) (\d+)/) {
		$time = $3;
		my $date = "$2/$1/$4"; $date =~ s/^ /0/;
		if (!defined $self->{lastDate} || $self->{lastDate} ne $date) {
			$self->{lastDate} = $date;
			$premsg = "---- $date ----\n";
		}
	} else {
		$time = time;
	}
	$msg =~ s/^/$time /mg;

   open(FILE, ">>$fileName") || croak "Can't append to $fileName: [$!]\n";
   print FILE $premsg, $msg;
   close(FILE) || croak "Can't close $fileName: [$!]\n";
}

=head1 NAME

Rui::Session::Base - a basic session-aware Rui server

=head1 SYNOPSIS

This is an abstract class, subclasses should implement the B<run> method.

=head1 DESCRIPTION

This is a common Rui server implementation, that all real servers should
inherit from. The communication method with clients is left to the subclasses.

It handles client requests through a specific
I<Rui::Remote::Session>. Sessions are managed through a
I<Rui::Remote::SessionManager>.

=head1 METHODS

=head2 new

=over 4

=item description

Create a new server.

=item parameters

Hash of named parameters. Keys/values defined here:

I<LogFile> - Scalar. Path to log file. Set to "-" for dumping log to
stdout.

I<PidFile> - Scalar. Path to pid file.

=back

=cut

1;
