package Rui::Server::HTTP;
use strict;

use Carp;
use Attribute::Protected;
use Rui::Remote::ClientMessage;
use HTTP::Status;
use HTTP::Daemon;
use HTTP::Date qw(time2str);
use LWP::MediaTypes qw(guess_media_type);

# Perl 5.8 and later only
use open ':utf8';

use base 'Rui::Server::Base';

$SIG{'PIPE'} = 'IGNORE';

my $CRLF = "\015\012";

sub new {
	my ($this, %params) = @_;
	my $self = $this->SUPER::new(%params);
	$self->{isDebug}     = $params{IsDebug} || 0;
	$self->{mainRoot}    = "$Rui::PROJECTS_HOME/../web";
	$self->{baseRoot}    = "$Rui::PROJECTS_HOME/$Rui::PROJECT_BASE_NAME/web";
	$self->{projectRoot} = "$Rui::PROJECT_HOME/web"
		if $Rui::PROJECT_NAME ne $Rui::PROJECT_BASE_NAME;
	$self->{projectRoot} = delete $params{DocumentRoot}
		if $params{DocumentRoot};
	$self->{daemon}      = HTTP::Daemon->new(Reuse => 1, %params) ||
		croak 'Cannot start server';
	return $self;
}

sub run : Public {
	my $self = shift;
	$self->log(
		'Started Rui HTTP-based server for '. $self->{daemon}->url,
		'Development mode: '. $self->{isDebug},
		'Projects home: '. $Rui::PROJECTS_HOME,
		'Project name: '. $Rui::PROJECT_NAME,
		'Main root: '. $self->{mainRoot},
		'Base root: '. $self->{baseRoot},
		defined $self->{projectRoot}? 'Project root: '. $self->{projectRoot}: (),
		'Applications: '. join(', ', Rui::Application::Map->new->getApplicationNames),
		'------------------ Waiting for requests ------------------',
	);
	while (my $client = $self->{daemon}->accept) {
		my $request = $client->get_request;
		next unless $request;
		$self->handleRequest($client, $request);
		$client->close;
	}
}

sub handleRequest : Protected {
	my ($self, $client, $request) = @_;

	my $hostname = gethostbyaddr($client->peeraddr, AF_INET) || $client->peerhost;
	$self->log($hostname . " " . $request->uri);

	# loop for testing: responds with the contents of the request
	if ($request->uri =~ qr{/loop\?}) {
		$self->handleApplicationRequest($client, $request, 1);

	} elsif ($request->uri !~ qr{/rui(?:Perl)?\?}) {
		$self->handleFileRequest($client, $request);

	} else {
		$self->handleApplicationRequest($client, $request);
	}
}

sub handleApplicationRequest : Protected {
	my ($self, $client, $request, $isLoop) = @_;
	$isLoop ||= 0;

	# we get the events from the XML content of the message or from the URL
	my $requestContent = $request->content;
	my %event = ($requestContent =~ /<rui>/)?
		Rui::Remote::ClientMessage->new($requestContent)->getEvent:
		$request->uri->query_form;

	my $content = $isLoop?
   	join(', ', %event):
		eval { $self->handleApplicationEvent(\%event) } || "Server Error: $@";

# 	unless (defined $content) {
# 		my $errorMsg = $@ || $! || "Unknown error";
# 		$self->log("handleApplicationEvent error:\n" . $errorMsg);
# 		$errorMsg =~ s/\n/<br>\n/sg;
# 		$client->send_error(RC_EXPECTATION_FAILED, $errorMsg);
# 		return;
# 	}

	$client->send_basic_header;
	print $client
		"Content-Type: text/xml$CRLF".
		"Content-Length: ". length($content)."$CRLF".
		"Last-Modified: ", time2str(time), "$CRLF".
		"Expires: ", time2str(time), "$CRLF".
		"Connection: close$CRLF".
		$CRLF. $content;
}

sub handleFileRequest : Protected {
	my ($self, $client, $request) = @_;
	my $path = $request->uri->path;
	if ($path =~ qw{^/WEB-INF}) {
		$client->send_error(RC_FORBIDDEN);
		return;
	}
	$path .= '/index.html' if ($path =~ qr{/$});

	# file may be in the main web dir, the base dir, or in the project dir
	my $projectPath = $self->{projectRoot}? "$self->{projectRoot}/$path": undef;
	my $basePath    = "$self->{baseRoot}/$path";
	my $mainPath    = "$self->{mainRoot}/$path";

	my $realPath = 
		$projectPath && -f $projectPath? $projectPath:
		-f $basePath?                    $basePath:
		-f $mainPath?                    $mainPath:
		undef;

	if ($realPath) {
		$self->send_file_response($client, $realPath);
	} else {
		$client->send_error(RC_NOT_FOUND);
	}
}

sub send_file_response
{
	my($self, $client, $file) = @_;
	if (-d $file) {
		$client->send_dir($file);
	} elsif (-f _) {
		local(*F);
		sysopen(F, $file, 0) or return $client->send_error(RC_FORBIDDEN);
		binmode(F);
		my($ct,$ce) = guess_media_type($file);
		my($size,$mtime) = (stat _)[7,9];
		$client->send_basic_header;
		print $client "Content-Type: $ct$CRLF";
		print $client "Content-Encoding: $ce$CRLF" if $ce;
		print $client "Content-Length: $size$CRLF" if $size;
		print $client "Last-Modified: ", time2str($mtime), "$CRLF" if $mtime;
		if ($self->{isDebug}) {
			print $client "Expires: ", time2str(time-1), "$CRLF";
		} else {
			print $client "Expires: ", time2str(time + 60 * 60 * 24 * 5), "$CRLF" if $mtime;
		}
		print $client "Connection: close$CRLF";
		print $client $CRLF;
		$client->send_file(\*F);
	} else {
		$client->send_error(RC_NOT_FOUND);
	}
}

=head1 NAME

Rui::Server::HTTP - a session-aware HTTP-based Rui server

=head1 SYNOPSIS

  use Rui::Server::HTTP;

  Rui::Server::HTTP->new(
     LogFile   => "server.log",  # "-" for stdout
     LocalPort => 8888,
  )->run;

=head1 SUPERCLASS

L<Rui::Server::Base>

=head1 REQUIRES

L<Rui::Remote::ClientMessage>

=head1 DESCRIPTION

A subclass of L<Rui::Server::Base>. This server implementation uses
HTTP::Daemon to communicate with clients. Intended for web-browser based
clients.

=head1 METHODS

=head2 new

=over 4

=item description

Create a new server. Inherits all initialization parameters of
I<Rui::Server::Base>, I<HTTP::Daemon> and I<IO::Socket::INET>.

New named parameters:

I<DocumentRoot> - overrides the project's document root that defaults
to "$Rui::PROJECT_HOME/web".

=item parameters

Hash of named parameters. These are passed to L<Rui::Server::Base> and
to C<HTTP::Daemon>.

=back

=cut

1;
