package Rui::View::Base;

use Carp;
use Attribute::Protected;
use Attribute::Abstract;
use Rui::Model::Value;

use base 'Rui::Widget::Panel';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	croak 'adapt+model are exclusive' if $params{adapt} && $params{model};
	$self->{modelHolder} = Rui::Model::Value->new(
		$params{adapt}? (
			adapt           => $params{adapt},
			attachToAdaptee => 1,
		):(
			value           => $params{model} || $self->getDefaultModel,
		),
	);
	$self->{modelDefined} = 1;
	$self->SUPER::init(%params);
	$self->sync;
}

# accessing ----------------------------------------------------------------

sub model
{
	my ($self, $newModel) = @_;
	my $oldModel = $self->modelHolder->value;
	return $oldModel if @_ == 1;
	$self->modelHolder->value($newModel);
	return $self;
}

sub modelHolder { shift->{modelHolder} }

sub makeChildModelAdapter
{
	my ($self, $name) = @_;
	return Rui::Model::Value->new(attachToAdaptee => 1, adapt =>
		["value.getChildModel_$name" => $self->modelHolder]
	);
}

# compositing --------------------------------------------------------------

sub createWidget
{
	my ($self, %params) = @_;
	croak 'holder+adapt are exclusive' if $params{holder} && $params{adapt};
	$params{adapt} = [
		"value.getChildModel_$params{holder}" => $self->modelHolder
	] if $params{holder};
	return $self->SUPER::createWidget(%params);
}

# client handling ----------------------------------------------------------

sub sync : Protected
{
	my $self = shift;
	if (defined $self->model) {
		$self->sendToClient(disabled => 0)
			unless $self->disabled || $self->{modelDefined};
		$self->{modelDefined} = 1;
		$self->sendModelToClient;
	} else {
		$self->onUndefModel;
		$self->sendToClient(disabled => 1)
			unless $self->disabled || !($self->{modelDefined});
		$self->{modelDefined} = 0;
	}
	return  $self;
}

# event handling -----------------------------------------------------------

sub handleEvent_modelHolder_Change { shift->sync }

# template methods ---------------------------------------------------------

sub sendModelToClient  : Protected {}   # non composites will override this
sub getModelEventNames : Protected { () }
sub getDefaultModel    : Protected { undef } 
sub onUndefModel       : Protected {}   # non composites will clear() here

sub getAttachments : Protected
{
	my $self = shift;
	return {
		modelHolder => {
			Change => 'Change',
			map {( "Value_$_" => $_ )} $self->getModelEventNames,
		},
	};
}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::View::Base - view abstract base

=head1 SYNOPSIS

  # using views

  $view = $parentWidget->MyView; # use default model
  $view = $parentWidget->MyView(model => $model);
  $view->model($model);
  $model = $view->model;

  
  # subclassing: an on/off view

  package MyViews::OnOff;
  use Rui::Model::Value;
  use base 'Rui::View::Base';

  sub sendModelToClient
  {
     my $self = shift;
     $self->attribute(value => $self->model->value? 'On': 'Off');
  }

  sub getDefaultModel       { Rui::Model::Value->new }
  sub getDefaultWidgetClass { 'Label' }
  sub getModelEventNames    { qw(Change) }
  sub onUndefModel          { shift->clear }   

=head1 SUPERCLASS

L<Rui::Widget::Base>

=head1 REQUIRES

L<Rui::Model::Base>

=head1 DESCRIPTION

A view is just a widget, but has some model. If the model is undefined,
then the view is disabled. The view listens to model events, and
modifies its display accordingly. It listens to events from the user,
and modifies its model accordingly.

You can use C<model()> to get/set the model of the view, and you can set
it as an init key to the constructor. If created without a model, the
view will create its own default model.

The view uses a L<Rui::Model::Value> as a holder for the model. By
setting the C<adapt> key in the init hash, you can configure the view so
that its model changes automatically whenever some event propagates
through the value net. This saves you typing as there is no need to
write the event handler that replaces the model, and the attach/detach
mechanism for that event handler.

=head1 SUBCLASSING

You may implement the following template methods:

=over 4

=item *

C<populate> - Will be called when the view is created. Composite views
want to create their children here. Non-composite views will leave this
empty.

=item *

C<sendModelToClient> - Will be called when the model is replaced, or
when the model fires a C<Change> event. Composite views will do nothing
here, as the value net will propagate to the sub views pieces of the
parent view model automatically. Non-composite views will want to clear
the contents of their client-side widget, and render the model to it.
Non-composites will want to call C<sync()> when there is a large change
in their model that requires a complete refresh. It calls
C<sendModelToClient()> if the view model is defined, or disables the
view if not. You should not call C<sendModelToClient()>.

=item *

C<getDefaultModel> - Should return an object that can be used as a
default model for the view, if none was specified in the constructor.
If you do not write this method the model will be set to undef.

=item * 

C<getModelEventNames> - Should return an array of strings. These are
event names on the model, that are deemed interesting by this view. The
framework will add/remove these listeners on the model, automatically
whenever there is a new model. You just need to say what events you need
from the model, and provide event handlers. If your view is a composite,
and its children listen to parts of its model, then there is probably no
need for the composite to implement this method, as it does not need to
listen to its model.

=item * 

C<onUndefModel> - Non-composites will want to C<clear()> themselves
here. Default implementation does nothing.

=back

You must implement the following template methods:

=over 4

=item *

C<handleEvent_modelHolder_SomeModelEventName> - for every event name you
put in C<getModelEventNames>, you must add the corresponding event
handler for that model event. Here you want to render any changes from
the model. You do not need to write add a handler for one event:
C<Change>. The framework will automatically call C<sendModelToClient()>
when this event is called.

=back

=head1 COMPOSITING

Many times a composite view will want to give its children some part of
its own model: in MVC there are parallel hierarchies of views and
models. This means composite views need to replace the models of their
child views when their own model is replaced. The framework does this
for you if you do two things:

=over 4

=item 1

Create the child views that you want to attach to parts of your model
using the C<holder> C<init()> key, I<not> the C<model> key. You set the
value to the name of the child model of the model of the composite view,
that you want to give to the child view. Example:
L<Rui::View::DoubleList> creates a child list view on a part of its
model so:

  $self->List(holder => 'left');

Instead of:

  $self->List(model => $self->model->left);

Now that the framework knows how to get the holder for the child model
from the composite model, it can replace the models of the children
views automatically when the composite view model is replaced.

If the view you are creating is not a direct child of this view, you can
use C<makeChildModelAdapter()> to get an adapter on the child model.

=item 2

Add methods to the model of your composite view that return named child
models that you want to give to the children views. Example:
L<Rui::Model::DoubleList> has a method C<getChildModel_left()> that
returns the selection in list that models the left list. The framework
will call this method to get the new model for the child view when the
model of the composite view is replaced.

=back

=head1 METHODS

=head2 ACCESSING

=head3 model

=over 4

=item parameters

Setting: model - L<Rui::Model::Base>. A new model.

=item description

Get/set model. View is refreshed when model changes.

=back

=head3 modelHolder

=over 4

=item returns

L<Rui::Model::Value>. The model holder.

=item description

The model holder is immutable for the view.

=back

=head3 makeChildModelAdapter

=over 4

=item parameters

Scalar. The name of the child model of the model of this view.

=item returns

L<Rui::Model::Value>. An adapter on the child model.

=item description

Creates a new value model as an adapter on a child model of this view.

=back

=cut