package Rui::View::List;

use Carp;
use Attribute::Protected;
use Rui::Model::SelectionInList;

use base 'Rui::View::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	$self->{renderer} = $params{renderer} || $self;
	$self->SUPER::init(%params);
}

sub destroy
{
	my ($self, $dontSendToClient) = @_;
	delete $self->{renderer};
	$self->SUPER::destroy($dontSendToClient);
}

# event handling -----------------------------------------------------------

sub handleEvent_modelHolder_Selection_Change { shift->sendSelectionToClient }
sub handleEvent_modelHolder_List_Change      { shift->sync }

sub handleEvent_modelHolder_List_RangeAdded
{
	my ($self, $event) = @_;
	my ($index, $size) = map { $event->$_ } qw(index size);
	for (my $i = $index; $i < $index + $size; $i++)
		{ $self->sendItemAddedToClient($i) }
}

sub handleEvent_modelHolder_List_RangeRemoved
{
	my ($self, $event) = @_;
	my ($index, $size) = map { $event->$_ } qw(index size);
	for (my $i = $index; $i < $index + $size; $i++)
		{ $self->sendItemRemovedToClient($index) }
}

sub handleEvent_modelHolder_List_RangeChanged
{
	my ($self, $event) = @_;
	my ($index, $size) = map { $event->$_ } qw(index size);
	my $selectedItem   = $self->model->selectedItem;
	my $selectedIndex  = defined $selectedItem? $self->model->selectedIndex: -1;
	for (my $i = $index; $i < $index + $size; $i++) {
		# should be: sendItemChangedToClient
		$self->sendItemRemovedToClient($i);
		$self->sendItemAddedToClient($i);
		$self->sendSelectionToClient if $selectedIndex == $index;
	}
}

# client handling ----------------------------------------------------------

sub sendModelToClient : Protected
{
	my $self = shift;
	$self->clear;
	my $list = $self->model->list;
	return unless $list && $list->getSize;
	for (my $i=0; $i < $list->getSize; $i++)
		{ $self->sendItemAddedToClient($i) }
	$self->sendSelectionToClient if defined $self->model->selectedItem;
}

sub sendSelectionToClient : Private
{
	my $self = shift;
	my $selectedItem = $self->model->selectedItem;
	$self->attribute(selectedIndex =>
		defined $selectedItem && $selectedItem ne ''?
			$self->model->selectedIndex: -1
	);
}

sub sendItemAddedToClient : Protected
{
	my ($self, $index) = @_;
	my $item = $self->model->list->get($index);

	my $rendering = ref $self->{renderer} eq 'CODE'?
		&{$self->{renderer}}($item):
		$self->{renderer}->render($item);
	$self->sendToClient(insert => $index , $rendering->{value});
	$rendering->{image}->setOn($self, setItemImage => $index) if
		exists $rendering->{image};
}

sub sendItemRemovedToClient : Protected
{
	my ($self, $index) = @_;
	$self->attribute(removeByIndex => $index);
}

sub handleClientEvent_SelectionChanged
{
	my ($self, %eventData) = @_;
	my $selectedIndex = $eventData{selectedIndex};
	croak "client sent undefined selected index"
		unless defined $selectedIndex;
	# die unless selectedIndex numeric
	$self->model->selectedIndex($selectedIndex);
}

# rendering ----------------------------------------------------------------

sub render
{
	my ($self, $item) = @_;
	return { value =>
		UNIVERSAL::can($item, 'value')?
			$item->value:
		ref $item eq 'HASH' && exists $item->{value}?
			$item->{value}:
			scalar $item
	};
}

# template methods ---------------------------------------------------------

sub getDefaultWidgetClass : Protected { 'List' }
sub getDefaultModel       : Protected { Rui::Model::SelectionInList->new }
sub onUndefModel          : Protected { shift->clear }   
sub getModelEventNames    : Protected
{qw(
	List_Change
	List_RangeAdded
	List_RangeRemoved
	List_RangeChanged
	Selection_Change
)}

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::View::List - the list view

=head1 SYNOPSIS

  $list = $parent->List;                               # default renderer
  $list = $parent->List(renderer => $self);            # call $self->render($item)
  $list = $parent->
     List(renderer => sub { {value => shift->name} }); # callback renderer

=head1 SUPERCLASS

L<Rui::View::Base>

=head1 MODEL

L<Rui::Model::SelectionInList>

=head1 DESCRIPTION

A list view work with a L<selection in list
model|Rui::Model::SelectionInList>. It paints any changes from the list
and the selection of the model, and changes the selection model when
users modify the selection using the mouse or keyboard.

The list uses a renderer to render each item in the list. A renderer is
a callback or an object, in which case the method C<render()> will be
called.

The renderer recieves one parameter: the item currently to be rendered
by the view.

The renderer must return a rendering. A rendering is a hash ref. It will
be used to construct a label for the list item. Valid keys are C<value>,
and C<image>.

The list view has a default renderer that tries to DWIM.

=cut
