package Rui::Widget::Base;

use Carp;
use Attribute::Protected;
use Attribute::Abstract;

use base 'Rui::Event::Listenable';

# creating------------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	$self->SUPER::init(%params);

	$self->{parent}        = $params{parent} || croak 'No parent';
	$self->{window}        = $self->{parent}->getWindow;
	$self->{session}       = $self->{window}->getSession;
	$self->{widgetFactory} = $self->{window}->getWidgetFactory;
	$self->{windowId}      = $self->{parent}->getWindowId;
	$self->{widgetId}      = $self->{session}->makeWidgetId;
	$self->{widgetClass}   = $params{widgetClass} ||
		$self->getDefaultWidgetClass;

	$self->{session}->setClientListener($self->{widgetId}, $self);
	$self->sendToClient(create => $self->{parent}->getWidgetId);

	foreach (qw(style layoutHint disabled tooltip cssClass backgroundImage))
		{ $self->$_($params{$_}) if exists $params{$_} }
}

sub Window { shift->getWindow->Window(@_) }

sub destroy
{
	my ($self, $dontSendToClient) = @_;
	unless ($dontSendToClient) {
		$self->getParent->removeWidget($self);
		$self->sendToClient('destroy');
	}
	$self->{session}->removeClientListener($self->{widgetId});
	# should we not delete session?
	delete $self->{$_} foreach qw(window parent events);
}

# accessing ----------------------------------------------------------------

sub cssClass         { shift->attribute(cssClass => @_) }
sub tooltip          { shift->attribute(tooltip  => @_) }
sub disabled         { shift->attribute(disabled => @_) }
sub style            { shift->complexAttribute(style => @_) }
sub clear            { shift->sendToClient('clear') }
sub getWidgetId      { shift->{widgetId} }
sub getWindowId      { shift->{windowId} }
sub getWidgetFactory { shift->{widgetFactory} }
sub getWindow        { shift->{window} }
sub getParent        { shift->{parent} }
sub focus            { shift->sendToClient('focus') }

sub backgroundImage
{ 
	my ($self, $backgroundImage) = @_;
	return $self->{backgroundImage} if @_ == 1;

	$self->{backgroundImage}->removeFromBackground($self)
		if defined $self->{backgroundImage} && !defined $backgroundImage;

	$self->{backgroundImage} = $backgroundImage;
	$backgroundImage->addToBackground($self)
		if defined $backgroundImage;
	return $self;
}

sub layoutHint
{
	my ($self, @params) = @_;
	croak 'Missing all parameters' if @_ == 1 || !$params[0];
	return $self->complexAttribute(layoutHint => @params)
		if @params == 1 && !ref($params[0]);
	my %params = @params == 1? %{$params[0]}: @params;
	# this little DWIMer should be on the client
	if ($params{fill})
		{ $params{"${_}Align"} = 'fill' foreach qw(v h) }
	if ($params{center})
		{ $params{"${_}Align"} = 'center' foreach qw(v h) }
	if ($params{flex})
		{ $params{"${_}Flex"} = 1 foreach qw(v h) }
	return $self->complexAttribute(layoutHint => \%params);
}

sub attribute : Protected
{
	my ($self, $key, $value) = @_;
	return $self->{$key} if @_ == 2;
	$self->{$key} = $value;
	$self->sendToClient($key => $value);
	return $self;
}

sub complexAttribute : Protected
{
	# scary getter/setter, due to DWIMerness
	my ($self, $attributeName, @params) = @_;
	croak 'Missing all parameters' if @_ == 2 || !$params[0];
	if (@params == 1 && !ref($params[0])) {
		my $key = $params[0];
		croak "Not a valid key [$key]" unless $key;
		return $self->{$attributeName}->{$key};
	}

	my %params = @params == 1? %{$params[0]}: @params;
	foreach my $key (sort keys %params) {
		my $value = $params{$key};
		croak "Not a valid value. Key is [$key], value is [$value]"
			unless defined $value;
		$self->{$attributeName}->{$key} = defined $value? $value: '';
		$self->sendToClient($attributeName => $key, $value);
	}
	return $self;
}

# template methods ---------------------------------------------------------

sub getDefaultWidgetClass : Abstract;

# event handling -----------------------------------------------------------

sub handleClientEvent
{
	my ($self, %params) = @_;
	my $handler = "handleClientEvent_$params{name}";
	$self->$handler(%params);
}

# remoting -----------------------------------------------------------------

sub sendToClient : Protected
{
	my ($self, $methodName, @params) = @_;
	$self->{session}->addWidgetCommand(
		(map {$self->{$_}} qw(widgetClass windowId widgetId)),
		$methodName, @params,
	);
	return $self;
}

# debugging ----------------------------------------------------------------

sub showMarkup
{
	my ($self, $resultWidget) = @_;
	$self->sendToClient('showMarkup' =>
		$resultWidget->getWindowId, $resultWidget->getWidgetId
	);
}

sub monitorServerMessages { shift->sendToClient('monitorServerMessages') }

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Widget::Base - abstract widget base class

=head1 SYNOPSIS

  # using widgets
  
  use MyWidgets::WhiteButton;
  # you can create widgets in windows or composites
  $window      = Rui::Window->new;
  $widget      = $window->Panel->MyWidgets__WhiteButton;
  $newWidget   = $window->Panel->MyWidgets__WhiteButton(
     value           => 'click to change value',
     style           => {padding => '2px 3px 1px 2px', color => '#FF0000'},
     layoutHint      => {minHeight => 100, minWidth => 50},
     disabled        => 0, # this is default: widget is enabled
     tooltip         => 'foo',
     events          => {Click => sub { shift->source->value('foo') }},
     cssClass        => 'MyButtonStyleClass',
     backgroundImage => $image,
  );
  # you can create windows from widgets
  $newWindow = $widget->Window(width => 200, height => 200);
  
  $widget->style(fontWeight => 'bold', color => '#00FF00');
  $widget->cssClass('MyButtonStyleClass');
  $widget->backgroundImage($image);
  $widget->disabled(1);
  $widget->tooltip('bar');
  $widget->layoutHint(hAlign => 'fill');
  print $widget->tooltip;
  print $widget->style('color');
  print $widget->cssClass;
  print $widget->backgroundImage;
  $widget->addListener(Click => $someListener);
  $widget->removeListener(Click => $someListener);
  $widget->destroy;


  # subclassing: a white button widget
  
  package MyWidgets::WhiteButton;
  use Rui::Event;
  use base 'Rui::Widget::Base';

  sub init
  {
     my ($self, %params) = @_;
     $self->SUPER::init(%params);
     $self->value($params{value}) if exists $params{value};
     $self->sendToClient(style => 'backgroundColor', '#FFFFFF');
  }

  sub value { shift->attribute(value => @_) }

  sub handleClientEvent_Click : Protected
  {
     my ($self, %params) = @_;
     # widgets are listenable so you can fire events
     $self->fireEvent(Rui::Event->new(
        name   => 'Click',
        source => $self,
     );
  }

  sub getDefaultWidgetClass { 'Button' }
  
=head1 SUPERCLASS

L<Rui::Event::Listenable>

=head1 DESCRIPTION

A proxy on some widget running on the client. The server side of the
widget. Supports sending arbitrary messages to subject, and listening to
subject events.

You create widgets by calling a method named after their type.

  $button = $parent->Button(value => 'foo');

Will create a button widget inside the parent widget. The widget
constructor takes an optional hash of named parameters, so you can
configure everything about the widget (except its children) at once.

Widgets are destroyed with C<destroy()>.

Widgets are styled using the CSS standard:

  $button->style(backgroundColor => '#FF00FF');
  $button->cssClass('MyButtonClass');

You can configure layout hints for the widget. These are used by the
layout manager of the parent. E.g. If the parent was configured with a
grid layout manager, then this could be done:

  $button->layoutHint(flex => 1);

Different layout managers will look for different hints. See
L<Rui::Widget::Panel> for possible values.

Widgets collaborate with a L<Rui::Remote::Session>. The session is
used for communication with the client side of the widget: appending to
the session buffer, and receiving client events. It is also used to get
the next available widget id.

Widgets collaborate with a L<Rui::Widget::Factory>, which they get from
their parent. They use it to create child widgets. They get it from
their window, which they get from their parent.

=head1 CSS

These are the style keys that you can use when styling widgets. See the
W3C CSS documentation for more information. Note that you cannot set
position or sizing keys (e.g. C<height>, C<top>), as these are managed
by the layout manager.

=over 4

=item font, textDecoration, letterSpacing

You can also use more specific keys: C<fontFamily>, C<fontSize>,
C<fontVariant>, C<fontWeight>.

=item backgroundAttachment, backgroundPosition, backgroundRepeat 

These effect how the background image is shown.

=item backgroundColor, color 

Color must be set as a hex RGB value. Here we set the color of the label
text to blue:

  $label->color('#0000FF');

=item border

You can specify C<borderBottom>, C<borderBottomColor>, etc. to
change specific attributes of a border. You must specify width in
pixels, by adding C<px> to the value. Here we set the left border width
to 2 pixels:

  $widget->style(borderLeftWidth => '2px');

=item margin, padding

You can access specific paddings and margins using C<marginBottom>,
C<paddingLeft>, etc. You must specify width in pixels, by adding C<px>
to the value, as with borders.

=item overflow, overflowX, overflowY 

These determine whether the widget content is hidden or can be scrolled,
when it exceeds the size of the widget. There are 3 options: C<hidden>
(default), C<auto>, and C<scroll>. You cannot set overflow to
C<visible>.

=item cursor 

=back

=head1 SUBCLASSING

You may have to create a new client-side peer for your new widget, if
the peer of the superclass is not good enough.

You can set attributes or complex attributes using C<attribute()>, or
C<complexAttribute()>.

Sending messages to the client-side peer is done through
C<sendToClient()>.

Overriding:

=over 4

=item *

C<handleClientEvent_SomeClientEventName> - will be called with event
data hash as only parameter whenever the client side of this widget
fires an event. The name of the event determines the method name called.
The name of the event is decided on the client side of the widget, where
the event originates.

=item *

C<getDefaultWidgetClass> - must return scalar name of the client side class
that will be used for this widget.

=back

If you create your own widgets, they will NOT be mapped to nice short
names like the Rui classes. For example a L<Rui::Widget::Label>, can be
created using C<< $parent->Label >>, because there is a type-to-class
mapping in the widget factory, from C<Label> to C<Rui::Widget::Label>.
Your class does not have a mapping there, so it will not work.

In this case you construct the type from the class name by going
C<s/::/__/g> on the class name to get the type. E.g. if your class is
C<Buttons::GreenButton>, you could instantiate it with:

  $greenButton = $parent->Buttons__GreenButton(value => 'foo');

You can add type-to-class mappings by subclassing the L<widget
factory|Rui::Widget::Factory>.

=head1 METHODS

=head2 CREATING

=head3 Widget type name, or class name after C<s/::/__/g>

=over 4

=item parameters

Hash of optional named parameters. Every method can be called by using
the method name as a key in the constructor hash.

=item returns

New widget.

=item description

You create widgets by calling a method, named after their class name, on
some composite. The following 2 lines are equivalent:

  $parent->Label(value => 'foo');
  $parent->Rui__Widget__Label(value => 'foo');

Creating widgets by type (vs. by class) is done by the L<widget
factory|Rui::Widget::Factory>, according to its mapping.

=back

=head3 destroy

=over 4

=item description

Widget and children will be destroyed.

=back

=head3 Window

=over4

=item parameters

Hash of named parameters passed to the window constructor.

=item returns

L<Rui::Window>. The new window.

=item description

Window constructor.

=back

=head2 STYLING

=head3 style

=over4

=item parameters

Getting: key - Scalar. CSS style key.

Setting: style - Hash or hash ref. CSS key/value pairs.

=item description

Get/set style values. You can only get one at a time, but you can set
several together. See CSS guide for possible values.

NOTE: do not set size and position keys, except padding/border/margin,
because these are set by the parent layout manager.

=back

=head3 cssClass

=over4

=item parameters

Setting: CSS class - Scalar. Name of CSS class to add to widget.

=item description

Add a CSS class to the widget. Instead of setting specific style keys,
it is usually better to create style classes. You define the CSS class
in a C<.css> file, that you place at C<$Rui::PROJECT_HOME/web/style>.
You also need to reference the stylesheet from the launcher page.

=back

=head3 backgroundImage

=over4

=item parameters

Setting: image (L<Rui::Image>)

=item description

Set a background image on the widget. Set as undefined to remove it.

=back

=head3 layoutHint

=over4

=item parameters

Getting: key - Scalar. layout hint key.

Setting: style - Hash or hash ref. Layout hint key/value pairs.

=item description

Get/set layoutHint values. You can only get one at a time, but you can
set several together. Each layout manager defines different possible
layout hints for children. See L<Rui::Widget::Panel> for possible
values.

=back

=head2 ACCESSING

=head3 disabled

=over4

=item parameters

Setting: disabled - Boolean.

=item description

Get/set the disabled state of this widget. Disabled widgets will not
receive events from the user.

=back

=head3 tooltip

=over4

=item parameters

Setting: tooltip - Scalar.

=item description

Get/set tooltip.

=back

=cut
