
package Rui::Widget::Factory;

use Carp;
use Attribute::Protected;

my $Instance;

sub instance
{
	$Instance = Rui::Widget::Factory->new unless $Instance;
	return $Instance;
}

sub getTypeToClassMap
{ {

	Panel          => 'Rui::Widget::Panel',
	SplitPanel     => 'Rui::Widget::SplitPanel',
	Label          => 'Rui::Widget::Label',
	Button         => 'Rui::Widget::Button',
	Grid           => 'Rui::Widget::Grid',
	NoteBook       => 'Rui::Widget::NoteBook',
	Tab            => 'Rui::Widget::Tab',
	Group          => 'Rui::Widget::Group',
	TextField      => 'Rui::View::TextField',
	PasswordField  => 'Rui::View::PasswordField',
	TextArea       => 'Rui::View::TextArea',
	HTMLArea       => 'Rui::View::HTMLArea',
	HTMLEditor     => 'Rui::View::HTMLEditor',
	List           => 'Rui::View::List',
	DoubleList     => 'Rui::View::DoubleList',
	ComboBox       => 'Rui::View::ComboBox',
	Tree           => 'Rui::View::Tree',
	Table          => 'Rui::View::Table',
	PerlSource     => 'Rui::View::PerlSource',
	CheckBox       => 'Rui::View::CheckBox',

} }

sub new
{
	my ($thing, %params) = @_;
	my $self = bless ({}, ref($thing) || $thing);
	$self->init(%params);
	return $self;
}

sub init : Protected
{
	my ($self, %params) = @_;
	$self->{typeToClassMap} = $self->getTypeToClassMap;
	$self->{loadedClasses} = {};
}

sub createWidget
{
	my ($self, %params) = @_;
	my $widget;
	my $type  = $params{type} || croak 'No type given for new widget';
	my $class = $self->getClassByType($type);

	unless ($self->{loadedClasses}->{$class}) {
		eval "use $class";
		croak "Failed using $class: $@" if $@;
		$self->{loadedClasses}->{$class} = 1;
	}
	$widget = $class->new(%params);
	return $widget;
}

sub getClassByType : Protected
{
	my ($self, $type) = @_;
	return $type if $type =~ /::/;
	return $type if $type =~ s/__/::/g;
	return $self->{typeToClassMap}->{$type}
		if exists $self->{typeToClassMap}->{$type};
	croak "Widget type: [$type] does not map to a widget class";
}

1;

=head1 NAME

Rui::Widget::Factory - widget factory

=head1 SYNOPSIS

  $factory = Rui::Widget::Factory->new;

  $factory->createWidget(
     type   => 'Label',
     parent => $parent,
     value  => 'foo',
  );
  $factory->createWidget(
     type   => 'Rui__Widget__Label',
     parent => $parent,
     value  => 'foo',
  );
  $factory->createWidget(
     type   => 'Rui::Widget::Label',
     parent => $parent,
     value  => 'foo',
  );

=head1 DESCRIPTION

The widget factory does 2 things to ease the creation of widgets:

=over 4

=item *

C<use> the widget classes so you don't need to

=item *

map between types and classes

=back

It should only be used from inside the framework.

=cut