package Rui::Widget::Panel;

use Carp;
use Attribute::Protected;

use base 'Rui::Widget::Base';

# creating -----------------------------------------------------------------

sub init : Protected
{
	my ($self, %params) = @_;
	$self->SUPER::init(%params);
	$self->{children} = [];
	$self->layoutManager($params{layoutManager})
		if exists $params{layoutManager};
	$self->populate(%params);
}

sub destroy
{
	my ($self, $dontSendToClient) = @_;
	$self->destroyChildren;
	delete $self->{children};
	$self->SUPER::destroy($dontSendToClient);
}

# compositing --------------------------------------------------------------

sub createWidget
{
	my ($self, %params) = @_;
	my $widget = $self->getWidgetFactory->createWidget(
		%params, parent => $self,
	);
	# should obey index
	push @{$self->{children}}, $widget;
	return $widget;
}

sub removeWidget : Protected
{
	my ($self, $widget) = @_;
	my $index  = $self->computeIndex( $self->getIndexByWidget($widget) );
	splice(@{$self->{children}}, $index, 1);
}

sub getIndexByWidget
{
	my ($self, $widget) = @_;
	my $index = 0;
	my @children = @{$self->{children}};
	$index++ until $index > @children || $widget eq $children[$index];
	croak 'Widget not in this parent' unless $children[$index] eq $widget;
	return $index ;
}

sub getWidgetByIndex
{
	my ($self, $index) = @_;
	return $self->{children}->[$self->computeIndex($index)];
}

sub firstChild { shift->getWidgetByIndex(0) }
sub lastChild  { shift->getWidgetByIndex(-1) }
sub childCount { scalar @{shift->{children}} }

# applies some DWIMers to an index
sub computeIndex : Protected
{
	my ($self, $index, $isForCreate) = @_;
	$isForCreate = defined $isForCreate? $isForCreate? 1: 0: 0;
	my $size     = $self->childCount;
	croak "Index not defined" unless defined $index;
	croak "Index not numeric:[$index]"
		unless $index =~ /^[+-]?\d+$/;
	$index = ($size + $index + $isForCreate) if $index < 0;
	croak "Out of range index:[$index],size:[$size]" if
		($index < 0) || ($index > $size) || (!$isForCreate && $index == $size);
	return $index;
}

sub AUTOLOAD
{
	my $self = shift;
	return if $AUTOLOAD =~ /DESTROY$/;
	$AUTOLOAD =~ s/^.*:://;
	return $self->createWidget(type => $AUTOLOAD, @_);
}

# accessing ----------------------------------------------------------------

sub clear
{
	my $self = shift;
	$self->destroyChildren;
	$self->{children} = [];
	$self->SUPER::clear;
}

# layout -------------------------------------------------------------------

sub layoutManager { shift->complexAttribute(layoutManager => @_) }

# template methods ---------------------------------------------------------

sub destroyChildren : Protected
{
	my $self = shift;
	$_->destroy(1) foreach @{$self->{children}};
}

sub populate              : Protected {} # composites will override this     
sub getDefaultWidgetClass : Protected { 'Panel' }

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Widget::Panel - composite widget

=head1 SYNOPSIS

  $panel = $someWindow->Panel;
  $panel->layoutManager(name => 'grid', numCols => 3);
  $panel->Label(value => 'foo');

=head1 SUPERCLASS

L<Rui::Widget::Base>

=head1 DESCRIPTION

A Panel lets you create widgets inside it and lay them out according to
some layout manager.

=head1 LAYOUT

You do not set size or location explicitly for widgets, unless you are
using the rarely used C<null> layout manager. Instead you set a layout
manager on the parent composite, configure it, and configure any layout
hints on the child widget. The layout manager will layout the widget
according to its configuration and the hints of the children. Layout
will occur after any relevant change to the composite or to the
children, and when the user resizes the window. The layout algorithm
operates automatically, so there is no need to explicitly start layout.

Here is the list of layout managers, and the hints with which they can
be configured.

=head2 row

Children are layed out in a horizontal row.

Layout manager keys:

=over4

=item pack

Boolean. If true, children are sized at their minimum sizes. If false,
children are equally sized at the maximum of their minimum sizes.

=item justify

Boolean. If true, children are spread out to fill available space in the
composite. If false, they remain in their initial positions, as close to
each other as possible, left aligned.

=item spacing

Integer. Number of pixels to put between children.

=back

Default is pack=true, justify=false, spacing=3

Layout manager hints:

=over4

=item height, width

Integer. This size will be used on the child, instead of the layout
policy.

=back

=head2 box

Children are layed out in a in a box, horizontal or vertical. You can
set size and flex on children for the axis of the container orientation.
Children are fully stretched on the axis perpendicular to the container
orientation.

Layout manager keys:

=over4

=item orient

horizontal/vertical.

=back

Layout manager hints:

=over4

=item height, width

Integer. Integer. This size will be used on the child, overriding the
layout policy. You should only set the one parallel to the container
orientation.

=item flex

Integer. How greedy is this child for space?

=back

=head2 fill

Children fill the parent completely. Space is divided equally between
them.

Layout manager keys:

=over4

=item orient

horizontal/vertical.

=back

=head2 grid

Children are layed out in a grid.

Layout manager keys:

=over4

=item numCols

Integer. Number of columns in grid.

=item equalWidth

Boolean. If true, columns are given the same width, and stretched to
fill parent.

=item hSpacing, vSpacing

Integer. Number of pixels to put between cells.

=back

Default is numCols=1, equalWidth=false, hspacing=vspacing=5

Layout manager hints:

=over4

=item indent

Integer. Number of pixels to indent child from left edge of cell.

=item hAlign, vAlign

one of: begin, end, center, fill. Alignment of widget in cell.

=item hSpan, vSpan

Integer. Number of cells to span.

=item hFlex, vFlex

Boolean. If true widget grabs excess space in the specified direction,
i.e. it expands as much as possible.

=item height, width

Integer. This size will be used on the child, instead of the layout
policy.

=back

Default is hAlign=begin, vAlign=center, indent=0, hSpan=vSpan=1,
hFlex=vFlex=false

=cut
