package Rui::Window;

use Carp;
use Attribute::Protected;
use Rui::Event;
use Rui::Remote::SessionManager;
use Rui::Widget::Factory;

use base 'Rui::Event::Listenable';

# creating------------------------------------------------------------------

sub init
{
	my ($self, %params) = @_;
	$self->SUPER::init(%params);

	my $width       = $params{width}   || 640;
	my $height      = $params{height}  || 480;
	my $isResizable = exists $params{isResizable}? $params{isResizable}: 1;
	my $session     = $params{session} ||
		Rui::Remote::SessionManager->instance->getActiveSession;

	$self->{session}       = $session;
	$self->{windowId}      = $session->makeWindowId;
	$self->{application}   = $session->getApplication;
	$self->{widgetFactory} = $self->{application}->getWidgetFactory;
	$self->{windowClass}   = $self->getWindowClass;

	$self->{session}->setClientListener($self->{windowId}, $self);
	$self->{application}->registerWindow($self->{windowId}, $self);

	$self->sendToClient(create => $width, $height, $isResizable);
	$self->populate(%params);
	$self->title($params{title}) if exists $params{title};
}

sub populate : Protected
{
	my ($self, %params) = @_;
	$self->{widget} = $self->{widgetFactory}->
		createWidget(type => 'Panel', parent => $self, %params);
}

sub destroy
{
	my ($self, $dontSendToClient) = @_;
	$self->{widget}->destroy(1) if $self->{widget};
	$self->sendToClient('destroy') unless $dontSendToClient;
	$self->{session}->removeClientListener($self->{windowId});
	$self->{application}->unregisterWindow($self->{windowId});
	delete $self->{$_} foreach qw(widget session application);
}

# accessing ----------------------------------------------------------------

sub getWindow        { shift }
sub isAlive          { shift->{session}? 1: 0 }
sub getWidgetId      { -1 } # -1 is special windowId for widgets
sub getWindowId      { shift->{windowId} }
sub getSession       { shift->{session} }
sub getApplication   { shift->{application} }
sub getWidgetFactory { shift->{widgetFactory} }

sub title {
	my ($self, $title) = @_;
	return $self->{title} if @_== 1;
	$self->{title} = $title;
	$self->sendToClient(title => $title);
	return $self;
}

# event handling -----------------------------------------------------------

sub handleClientEvent
{
	my ($self, %params) = @_;
	my $handler = "handleClientEvent_$params{name}";
	$self->$handler(%params);
}

sub handleClientEvent_Close
{
	my ($self, %params) = @_;
	$self->fireEvent(Rui::Event->new(name => 'Close', source => $self));
	$self->destroy(1);
}

# remoting -----------------------------------------------------------------

sub sendToClient : Protected
{
	my ($self, $methodName, @params) = @_;
	$self->{session}->addWindowCommand
		($self->{windowClass}, $self->{windowId}, $methodName, @params);
}

# delegating ---------------------------------------------------------------

sub AUTOLOAD
{
	my $self = shift;
	return if $AUTOLOAD =~ /DESTROY$/;
	croak "unknown method $AUTOLOAD" unless $self->{widget};
	$AUTOLOAD =~ s/^.*:://;
	my $result = $self->{widget}->$AUTOLOAD(@_);
	# delegate needs to return self not delegate
	return
		ref($result) eq ref($self->{widget}) &&
		$result eq $self->{widget}?
			$self: $result;
}

# template methods ---------------------------------------------------------

sub getWindowClass : Protected { 'Window' }

# --------------------------------------------------------------------------

1;

=head1 NAME

Rui::Window - the Rui window

=head1 SYNOPSIS

  # using a window

  use Rui::Window;

  $window = Rui::Window->new(
     width       => 200,
     height      => 150,
     isResizable => 1,
     title       => 'Window Title',
  );
  $window->title('New Window Title);
  $window->Label(value => 'foo');
  $window->addListener
     (Close => sub { print "closed: ". shift->source->getWindowId });
  print $window->getWindowId;
  print $window->getWidgetId; # always -1, special id for windows
  $window->destroy;

  # subclassing a window: a hello world window

  package HelloWorldWindow;
  use base 'Rui::Window';
  sub init
  {
     my ($self, %params) = @_;
     $self->SUPER::init(%params);
     $self->Label(value => 'Hello World');
  }

=head1 REQUIRES

L<Rui::Event>, L<Rui::Widget::Factory>, Rui::Remote::SessionManager

=head1 DESCRIPTION

A window is a top level widget, at the root of the widget composition
tree. When creating a window you can configure C<width>, C<height>,
C<isResizable>, and the window C<title>.

The window delegates all widget calls to its main panel. 

The winodw registers/deregisters with a L<Rui::Application::Base>, and
uses it to get the widget factory.

=cut