
use("Widget_LayoutBuffer");
use("Widget_WindowManager");

Class("Widget_Base");

Widget_Base.trimWidthKeys  = [
	"paddingRight",     "paddingLeft",
	"borderRightWidth", "borderLeftWidth",
	"marginRight",      "marginLeft"
];
Widget_Base.trimHeightKeys = [
	"paddingTop",       "paddingBottom",
	"borderTopWidth",   "borderBottomWidth",
	"marginTop",        "marginBottom"
];

// creating -------------------------------------------------------------------

_.init = function (widgetId, parent, addArg1, addArg2) { try {
	this.isDisabled        = false;
	this.id                = widgetId;
	this.parent            = parent;
	this.window            = parent.getWindow();
	this.bubbleLayout      = this.defaultBubbleLayout();
	this.element           = this.createElement(this.getTag());
	this.element.className = this.getWidgetClass();
	this.element.id        = widgetId;
	this.element.me        = this;
	this.layout = {
		"isDirty": false,
		"hint"   : {},
		"cache"  : {}
	};
	if (!this.isSelectable()) this.element.onselectstart = this.nullHandler;
	this.populate();
	try { parent.add(this, addArg1, addArg2) } catch (e) { Throw(e,
		"Adding to parent:" + parent.$className + ":" + parent.getId()
	)}

} catch (e) { Throw(e,
	"While creating  " + this.$className + ":" +
	this.getTag() + ":" + this.getWidgetClass() + ":" + widgetId
)}}

_.destroy = function (dontPaint) { try {
	var element           = this.element;
	this.parent           = undefined;
	this.window           = undefined;
	this.element          = undefined;
	this.listeners        = undefined;
	element.me            = undefined;
	element.onselectstart = null;
	if (!dontPaint) element.parentElement.removeChild(element);

} catch (e) { Throw(e,
	"While destroying  " + this.$className + ":" +
	this.getTag() + ":" + this.getWidgetClass() + ":" + this.getId()
)}}

// accessing  -----------------------------------------------------------------

_.getId         = function () { return this.id }
_.getWindow     = function () { return this.window }
_.getIsDisabled = function () { return this.isDisabled }
_.focus         = function () { this.element.focus() }
_.show          = function () { this.element.runtimeStyle.display = "" }
_.hide          = function () { this.element.runtimeStyle.display = "none" }

_.tooltip  = function (value) { this.element.title = value }

_.isChildOf = function (parent)
	{ return parent.element.contains(this.element) }

_.disabled = function (value) {
	value = (value && value != "0")? true: false;
	this.element.disabled = value;
	this.isDisabled       = value;
	this.disableChildren(value);
}

// style  ---------------------------------------------------------------------

_.cssClass = function (value) { this.element.className += " " + value }

_.style = function (key, value) {
	this.element.style[key] = value;
	this.layoutChanged();
}

// layout ---------------------------------------------------------------------

_.layoutHint = function (key, value) {
	this.layout.hint[key] = value == "0"? 0: value;
	this.layoutChanged();
}

_.layoutChanged = function () {
	if (!this.bubbleLayout || this.layout.isDirty) return;
	this.layout.isDirty = true;
	this.layout.cache   = {};
	this.parent.layoutChanged(this, Widget_LayoutBuffer.get());
}

_.setBounds = function (top, left, outerWidth, outerHeight) { try {
	if (
		!this.layout.isDirty &&
		this.compareBounds(top, left, outerWidth, outerHeight)
	) return;

	// hack! in hidden notebook pages sometimes currentStyle is empty
	if (!this.element.currentStyle) return;

	this.unclipTrimHeight();
	this.unclipTrimWidth();
	var trimWidth   = this.getTrimWidth();
	var trimHeight  = this.getTrimHeight();
	var innerWidth  = outerWidth  - trimWidth;
	var innerHeight = outerHeight - trimHeight;
	if (innerWidth < 0) {
		this.clipTrimWidth(0 - innerWidth);
		innerWidth = 0;
	}
	if (innerHeight < 0) {
		this.clipTrimHeight(0 - innerHeight);
		innerHeight = 0;
	}

	var cache = this.layout.cache;
	var style = this.element.style;

	try {
		if (innerWidth  != cache.innerWidth ) style.width  = innerWidth  + "px";
		if (innerHeight != cache.innerHeight) style.height = innerHeight + "px";
		if (top         != cache.top        ) style.top    = top         + "px";
		if (left        != cache.left       ) style.left   = left        + "px";
	} catch (e) {
		Throw(e, "Setting inner:"+top+":"+left+":"+innerWidth+":"+innerHeight);
	}
	cache.innerWidth  = parseInt(innerWidth);
	cache.innerHeight = parseInt(innerHeight);
	cache.outerWidth  = parseInt(outerWidth);
	cache.outerHeight = parseInt(outerHeight);
	cache.top         = parseInt(top);
	cache.left        = parseInt(left);
	this.setChildrenBounds();
	this.layout.isDirty = false;

} catch (e) {
	Throw(e, "Setting outer:"+top+":"+left+":"+outerWidth+":"+outerHeight);
}}

// trim clipping --------------------------------------------------------------

_.clipTrimByAxis = function (axis, missingSpace) {
	var isWidth = (axis == "width");
	var clip    = {};
	var style   = this.element.style;
	var keys    = Widget_Base[isWidth? "trimWidthKeys": "trimHeightKeys"];
	var key;
	var i; for (i in keys) { key = keys[i];
		clip[key]    = this.getTrimInfo(key);
		var newSpace = clip[key] - missingSpace;
		missingSpace = missingSpace - clip[key];
		if (newSpace < 0) newSpace = 0;
		style[key]   = newSpace + "px";
		this.layout.cache[key] = newSpace;
		if (missingSpace <= 0) break;
	}
	this.layout[isWidth? "trimWidthClip": "trimHeightClip"] = clip;
}

_.clipTrimWidth  = function (missingSpace)
	{ this.clipTrimByAxis("width", missingSpace) }

_.clipTrimHeight = function (missingSpace)
	{ this.clipTrimByAxis("height", missingSpace) }

_.unclipTrimByAxis = function (axis) {
	var isWidth = (axis == "width");
	var clip    = this.layout[isWidth? "trimWidthClip": "trimHeightClip"];
	if (!clip) return;
	var style   = this.element.style;
	var Keys    = Widget_Base[isWidth? "trimWidthKeys": "trimHeightKeys"];
	var aspect;
	var i; for (i in Keys) { aspect = Keys[i];
		var aspectClippedSize = clip[aspect];
		if (!isDefined(aspectClippedSize)) break;
		style[aspect] = aspectClippedSize + "px";
		this.layout.cache[aspect] = aspectClippedSize;
	}
	this.layout[isWidth? "trimWidthClip": "trimHeightClip"] = null;
}

_.unclipTrimWidth  = function () { this.unclipTrimByAxis("width")  }
_.unclipTrimHeight = function () { this.unclipTrimByAxis("height") }

// layout info -------------------------------------------------------------

_.computeMinWidth = function ()
	{ return this.window.computeMinWidth(this.element) }

_.computeMinHeight = function ()
	{ return this.window.computeMinHeight(this.element) }

_.getTrimInfo = function (key) {
	var space;
	var layout = this.layout;
	space = layout.cache[key];
	if (isDefined(space)) return space;
	var style = this.element.currentStyle;
	space = 0 + (parseInt(style[key]) || 0);
	layout.cache[key] = space;
	return space;
}

_.getPaddingLeft       = function () { return this.getTrimInfo("paddingLeft") }
_.getPaddingRight      = function () { return this.getTrimInfo("paddingRight") }
_.getPaddingTop        = function () { return this.getTrimInfo("paddingTop") }
_.getPaddingBottom     = function () { return this.getTrimInfo("paddingBottom") }
_.getMarginLeft        = function () { return this.getTrimInfo("marginLeft") }
_.getMarginRight       = function () { return this.getTrimInfo("marginRight") }
_.getMarginTop         = function () { return this.getTrimInfo("marginTop") }
_.getMarginBottom      = function () { return this.getTrimInfo("marginBottom") }
_.getBorderLeftWidth   = function () { return this.getTrimInfo("borderLeftWidth") }
_.getBorderRightWidth  = function () { return this.getTrimInfo("borderRightWidth") }
_.getBorderTopWidth    = function () { return this.getTrimInfo("borderTopWidth") }
_.getBorderBottomWidth = function () { return this.getTrimInfo("borderBottomWidth") }

_.getTrimLeft          = function () { return this.getPaddingLeft()   + this.getMarginLeft()   + this.getBorderLeftWidth() }
_.getTrimRight         = function () { return this.getPaddingRight()  + this.getMarginRight()  + this.getBorderRightWidth() }
_.getTrimTop           = function () { return this.getPaddingTop()    + this.getMarginTop()    + this.getBorderTopWidth() }
_.getTrimBottom        = function () { return this.getPaddingBottom() + this.getMarginBottom() + this.getBorderBottomWidth() }

_.getChildEdgeTop      = function () { return this.getPaddingTop() }
_.getChildEdgeLeft     = function () { return this.getPaddingLeft() }

_.getTrimWidth = function () {
	var space;
	var layout = this.layout;
	space = layout.cache.trimWidth;
	if (isDefined(space)) return space;
	space = 0 + this.getTrimLeft() + this.getTrimRight();
	layout.cache.trimWidth = space;
	return space;
}

_.getTrimHeight = function () {
	var space;
	var layout = this.layout;
	space = layout.cache.trimHeight;
	if (isDefined(space)) return space;
	var style = this.element.currentStyle;
	space = 0 + this.getTrimTop() + this.getTrimBottom();
	layout.cache.trimHeight = space;
	return space;
}

_.getOffsetLeft = function () {
	var size = 0;
	var element = this.element;
	while (element && element.tagName != "HTML") {
		size += element.offsetLeft + element.clientLeft;
		element = element.offsetParent;
	}
	size -= this.element.clientLeft + (parseInt(this.element.currentStyle.marginLeft) || 0);
	return size;
}

_.getOffsetTop = function () {
	var size = 0;
	var element = this.element;
	while (element && element.tagName != "HTML") {
		size += element.offsetTop + element.clientTop;
		element = element.offsetParent;
	}
	size -= this.element.offsetTop + (parseInt(this.element.currentStyle.marginTop) || 0);
	return size;
}

_.getMinWidth = function () {
	var space;
	var layout = this.layout;
	space = layout.hint.minWidth;
	if (isDefined(space)) return space;
	space = layout.cache.minWidth;
	if (isDefined(space)) return space;
	space = this.computeMinWidth();
	layout.cache.minWidth = space;
	return space;
}

_.getMinHeight = function () {
	var space;
	var layout = this.layout;
	space = layout.hint.minHeight;
	if (isDefined(space)) return space;
	space = layout.cache.minHeight;
	if (isDefined(space)) return space;
	space = this.computeMinHeight();
	layout.cache.minHeight = space;
	return space;
}

// event handling ----------------------------------------------------------

_.fireEvent = function (event) {
	if (this.getIsDisabled()) return;
	if (this.listeners && this.listeners[event.name]) {
		var listenerHash = this.listeners[event.name];
		listenerHash.method.call(
			listenerHash.listener,
			event,
			listenerHash.param1
		);
	} else {
		Client_Application.get().fireEvent(event);
	}
}

_.setListener = function (eventName, listener, method, param1) {
	if (!(this.listeners)) this.listeners = {};
	this.listeners[eventName]             = {};
	this.listeners[eventName].listener    = listener;
	this.listeners[eventName].method      = method;
	this.listeners[eventName].param1      = param1 || null;
}

_.nullHandler = function () {
	var event = this.document.parentWindow.event;
	event.returnValue  = false;
	event.cancelBubble = true;
	return false;
}

// utilities ---------------------------------------------------------------

_.createElement = function (tag) {
	var newElement = this.window.doc.createElement(tag);
	return newElement;
}

_.compareBounds = function (top, left, outerWidth, outerHeight) {
	var cache = this.layout.cache;
	if (outerWidth  != cache.outerWidth)  return false;
	if (outerHeight != cache.outerHeight) return false;
	if (top         != cache.top)         return false;
	if (left        != cache.left)        return false;
	return true;
}

_.clearElementChildren = function (element) {
	var child = element.firstChild;
	while (child) {
		nextChild = child.nextSibling;    
		element.removeChild(child);
		child = nextChild;
	}
}

// testing -----------------------------------------------------------------

_.serialize = function ()
	{ return "[" + this.element.className + ":" + this.element.id + "]" }

_.showMarkup = function (resultWindowId, resultWidgetId) {
	var result = this.window.getWidget(resultWindowId, resultWidgetId);
	var markup = this.element.outerHTML;
	result.value(markup);
	result.cssClass("MarkupArea");
	result.readOnly(true);
}

_.monitorServerMessages = function () {
	this.cssClass("MarkupArea");
	this.readOnly(true);
	Client_Runner.get().setMonitorArea(this);
}

// template methods --------------------------------------------------------

_.clear               = function () {}
_.populate            = function () {}
_.setChildrenBounds   = function () {}
_.disableChildren     = function () {}
_.defaultBubbleLayout = function () { return true }
_.getTag              = function () { Throw("Abstract method called: " + this.$className) }
_.getWidgetClass      = function () { Throw("Abstract method called: " + this.$className) }
_.isSelectable        = function () { return true }
