
Class("Widget_Layout_Grid");
Class_Singleton();

// creating -------------------------------------------------------------------

_.initLayout = function (widget) {
	var layout         = widget.layout;
	var manager        = layout.manager;
	manager.numCols    = parseInt(manager.numCols) || 1;
	manager.equalWidth = manager.equalWidth        || 0;
	manager.hSpacing   = manager.hSpacing          || 5;
	manager.vSpacing   = manager.vSpacing          || 5;
}

// layout ---------------------------------------------------------------------

_.computeMinWidth = function (widget) {
	if (!widget.hasChildren()) return 0;
	this.prepare(widget);
	var manager = widget.layout.manager;
	var minCols = widget.layout.cache.manager.minCols;
	var totalWidth = (minCols.length - 1) * manager.hSpacing;
	for (var i = 0; i < minCols.length; i++) totalWidth += minCols[i];
	return totalWidth;
}

_.computeMinHeight = function (widget) {
	if (!widget.hasChildren()) return 0;
	this.prepare(widget);
	var manager = widget.layout.manager;
	var minRows = widget.layout.cache.manager.minRows;
	var totalHeight = (minRows.length - 1) * manager.vSpacing;
	for (var i = 0; i < minRows.length; i++) totalHeight += minRows[i];
	return totalHeight;
}

_.setChildrenBounds = function (widget) {
	if (!widget.hasChildren()) return;
	this.prepare(widget);
	var totalWidth     = widget.getChildrenAreaWidth ();
	var totalHeight    = widget.getChildrenAreaHeight();
	if (!totalWidth || !totalHeight) return;
	var x              = widget.getChildEdgeLeft();
	var y              = widget.getChildEdgeTop ();
	if (y >= totalHeight || x >= totalWidth) return;
	var minWidth       = widget.getMinWidth () - widget.getTrimWidth ();
	var minHeight      = widget.getMinHeight() - widget.getTrimHeight();
	var manager        = widget.layout.manager;
	var cache          = widget.layout.cache.manager;
	var grid           = cache.grid;
	var numCols        = parseInt(manager.numCols);
	var numRows        = grid.length;
	var hSpacing       = manager.hSpacing, vSpacing       = manager.vSpacing;
	var minCols        = cache.minCols,    minRows        = cache.minRows;
	var flexCols       = cache.flexCols,   flexRows       = cache.flexRows;
	var flexColsLength = 0,                flexRowsLength = 0;
	var i, j;
	for (i in flexCols) { flexColsLength++; }
	for (i in flexRows) { flexRowsLength++; }
	var colWidths      = [].concat(minCols);
	var rowHeights     = [].concat(minRows);
	var extraWidth     = totalWidth  - minWidth;
	var extraHeight    = totalHeight - minHeight;
	if (flexColsLength) {
		var extra = parseInt(extraWidth / flexColsLength);
		var lastFlexCol = 0;
		for (i in flexCols) {
			colWidths[i] += extra;
			lastFlexCol = Math.max(lastFlexCol, i);
		}
		colWidths[lastFlexCol] += extraWidth % flexColsLength;
	}
	for (i in flexCols) {
		var flexColMinWidth = 0;
		for (j = 0; j < numRows; j++) {
			var child = grid[j][i];
			if (!child) continue;
			flexColMinWidth = Math.max(flexColMinWidth, child.layout.hint.width);
		}
		colWidths[i] = Math.max(flexColMinWidth, colWidths[i]);
	}
	if (flexRowsLength) {
		var extra = parseInt(extraHeight / flexRowsLength);
		var lastFlexRow = 0;
		for (i in flexRows) {
			rowHeights[i] += extra;
			lastFlexRow = Math.max(lastFlexRow, i);
		}
		rowHeights[lastFlexRow] += extraHeight % flexRowsLength;
	}
	for (i in flexRows) {
		var flexRowMinHeight = 0;
		for (j = 0; j < numCols; j++) {
			var child = grid[i][j];
			if (!child) continue;
			flexRowMinHeight = Math.max(flexRowMinHeight, child.layout.hint.height);
		}
		rowHeights[i] = Math.max(flexRowMinHeight, rowHeights[i]);
	}
	var equalWidth;
	if (manager.equalWidth) {
		equalWidth = parseInt
			((totalWidth - ((numCols - 1) * hSpacing)) / numCols);
		for (i = 0; i < numCols; i++) colWidths[i] = equalWidth;
	}
	for (var row = 0; row < numRows; row++) {
		var rowHeight = rowHeights[row];
		var gridRow   = grid[row];
		for (var col = 0; col < numCols; col++) {
			var child    = gridRow[col];
			var colWidth = colWidths[col];
			if (!child) {
				x += colWidth + hSpacing;
				continue;
			}
			var hint       = child.layout.hint;
			var spanWidth  = colWidth;
			for (i = 1; i < hint.hSpan; i++) {
				if ((col + i) > numCols) break;
				if (!manager.equalWidth) colWidth = colWidths[col + i];
				spanWidth += colWidth + hSpacing;
			}
			var spanHeight = rowHeight;
			for (i = 1; i < hint.vSpan; i++) {
				if ((row + i) > numRows) break;
				spanHeight += rowHeights[row + i] + vSpacing;
			}
			var childMinWidth = child.getMinWidth();
			var childX        = x + ((hint.hAlign == "center")?
				parseInt((spanWidth - childMinWidth)/2): (
					(hint.hAlign == "end")?
						spanWidth - childMinWidth - hint.indent: hint.indent
			));
			var childWidth = (hint.hAlign == "fill")?
				spanWidth - hint.indent: childMinWidth;
			var childMinHeight = child.getMinHeight();
			var childY    = y + ((hint.vAlign == "center")?
				parseInt((spanHeight - childMinHeight)/2):
					((hint.vAlign == "end")? spanHeight - childMinHeight: 0)
			);
			var childHeight = 
				parseInt(hint.vAlign == "fill"? spanHeight: childMinHeight) +
					hint.extraHeight;
			child.setBounds(childY, childX, childWidth, childHeight);
			x += colWidths[col] + hSpacing;
		}
		y += rowHeights[row] + vSpacing;
		x = widget.getChildEdgeLeft();
	}
}
// utility --------------------------------------------------------------------

_.prepare = function (widget) {
	var cache = widget.layout.cache;
	if (cache.manager && cache.manager.prepared) return;
	cache.manager = {};
	// create the grid
	this.loadWidgets(widget);
	// compute min sizes for cols, rows on non-span children
	this.computeNoSpanMin(widget);
	// compute flex cols, row
	this.computeFlex(widget);
	// compute min sizes for cols, rows for span children
	this.computeSpanMin(widget);
	// set as prepared
	cache.manager.prepared = true;
}

_.loadWidgets = function (widget) {
	var manager      = widget.layout.manager;
	manager.hSpacing = parseInt(manager.hSpacing);
	manager.vSpacing = parseInt(manager.vSpacing);
	manager.numCols  = parseInt(manager.numCols);
	var children     = widget.getChildren();
	var count        = children.length;
	var numCols      = manager.numCols;
	var grid         = [];
	var row          = 0, col = 0;
	grid.push(new Array(numCols));
	var i; for (i in children) { var child = children[i];
		var hint          = child.layout.hint;
		hint.indent       = parseInt(hint.indent) || 0;
		hint.width        = parseInt(hint.width)  || 0;
		hint.height       = parseInt(hint.height) || 0;
		hint.hSpan        = parseInt(hint.hSpan)  || 1;
		hint.vSpan        = parseInt(hint.vSpan)  || 1;
		hint.hFlex        = hint.hFlex            || 0;
		hint.vFlex        = hint.vFlex            || 0;
		hint.hAlign       = hint.hAlign           || "begin";
		hint.vAlign       = hint.vAlign           || "center";
		hint.extraHeight  = hint.extraHeight      || 0;
		while (grid[row][col] != undefined) {
			col++;
			if (col >= numCols) {
				row++;
				col = 0;
				if (row >= grid.length) grid.push(new Array(numCols));
			}
		}
		if (col + hint.hSpan - 1 >= numCols) {
			grid.push(new Array(numCols));
			row++;
			col = 0;
		}
		for (var j = 2; j <= hint.vSpan; j++) {
			if (row + j > grid.length) grid.push(new Array(numCols));
		}
		grid[row][col] = child;
		var rowFill = hint.vSpan - 1;
		var colFill = hint.hSpan - 1;
		for (var k = 1; k <= rowFill; k++)
			for (var l = 0; l < hint.hSpan; l++)
				grid[row + k][col + l] = 0;
		for (var k = 1; k <= colFill; k++)
			for (var l = 0; l < hint.vSpan; l++)
				grid[row + l][col + k] = 0;
		col += hint.hSpan - 1;
	}
	widget.layout.cache.manager.grid = grid;
}

_.computeNoSpanMin = function (widget) {
	var manager = widget.layout.manager;
	var cache   = widget.layout.cache.manager;
	var grid    = cache.grid;
	var numCols = manager.numCols,    numRows = grid.length;
	var minCols = new Array(numCols), minRows = new Array(numRows);
	var maxCol  = 0;
	for (var col = 0; col < numCols; col++) {
		var maxWidth = 0;
		for (var row = 0; row < numRows; row++) {
			var child = grid[row][col];
			if (!child || child.layout.hint.hSpan > 1) {
				minCols[col] = 0;
				continue;
			}
			var hint = child.layout.hint;
			maxWidth = Math.max(maxWidth,
				hint.indent + (hint.width || child.getMinWidth())
			);
		}
		minCols[col] = maxWidth;
		maxCol = Math.max(maxCol, maxWidth);
	}
	if (manager.equalWidth)
		for (var col = 0; col < numCols; col++) minCols[col] = maxCol;
	for (var row = 0; row < numRows; row++) {
		var maxHeight = 0;
		for (var col = 0; col < numCols; col++) {
			var child = grid[row][col];
			if (!child || child.layout.hint.vSpan > 1) {
				minRows[row] = 0;
				continue;
			}
			var hint = child.layout.hint;
			maxHeight = Math.max(maxHeight,
				(hint.height || child.getMinHeight())
			);
		}
		minRows[row] = maxHeight;
	}
	cache.minCols = minCols;
	cache.minRows = minRows;
}

_.computeFlex = function (widget) {
	var manager  = widget.layout.manager;
	var cache    = widget.layout.cache.manager;
	var grid     = cache.grid;
	var numCols  = manager.numCols, numRows  = grid.length;
	var flexCols = {},              flexRows = {};
	for (var col = 0; col < numCols; col++) {
		for (var row = 0; row < numRows; row++) {
			var child = grid[row][col];
			if (!child) continue;
			var hint  = child.layout.hint;
			if (hint.hFlex) flexCols[col + hint.hSpan - 1] = true;
			if (hint.vFlex) flexRows[row + hint.vSpan - 1] = true;
		}
	}
	cache.flexCols = flexCols;
	cache.flexRows = flexRows;
}

_.computeSpanMin = function (widget) {
	var manager  = widget.layout.manager;
	var cache    = widget.layout.cache.manager;
	var grid     = cache.grid;
	var numCols  = manager.numCols, numRows  = grid.length;
	var minCols  = cache.minCols,   minRows  = cache.minRows;
	var flexCols = cache.flexCols,  flexRows = cache.flexRows;
	for (var row = 0; row < numRows; row++) {
		for (var col = 0; col < numCols; col++) {
			var child = grid[row][col];
			if (!child) continue;
			var hint  = child.layout.hint;
			if (hint.hSpan > 1) {
				var childWidth = hint.width || child.getMinWidth();
				var spanWidth = 0 - hint.indent;
				for (var c = col; c < col + hint.hSpan; c++)
					spanWidth += minCols[c] + manager.hSpacing;
				spanWidth = spanWidth - manager.hSpacing;
				if (childWidth > spanWidth) {
					var missingSpace = childWidth - spanWidth;
					var lastCol = col + hint.hSpan - 1;
					if (manager.equalWidth) {
						var colExtra = parseInt(missingSpace / numCols);
						var i; for (i in minCols) minCols[i] += colExtra;
						minCols[lastCol] += missingSpace % numCols;
					} else {
						var localFlexCols = [];
						for (var i = col; i < lastCol; i++)
							if (flexCols[i]) localFlexCols.push(i);
						var localLength = localFlexCols.length;
						if (localFlexCols.length > 0) {
							var colExtra = parseInt(missingSpace / localLength);
							for (var j = 0; j < localLength; j++)
								minCols[ localFlexCols[j] ] += colExtra;
							minCols[lastCol] += missingSpace % localLength;
						} else {
							minCols[lastCol] += missingSpace;
			}}}}
			if (hint.vSpan > 1) {
				var childHeight = hint.height || child.getMinHeight();
				var spanHeight = 0;
				for (var r = row; r < row + hint.vSpan; r++)
					spanHeight += minRows[r] + manager.vSpacing;
				spanHeight = spanHeight - manager.vSpacing;
				if (childHeight > spanHeight) {
					var missingSpace = childHeight - spanHeight;
					var lastRow = row + hint.vSpan - 1;
					var localFlexRows = [];
					for (var j = row; j < lastRow; j++)
						if (flexRows[i]) localFlexRows.push(j);
					var localLength = localFlexRows.length;
					if (localLength > 0) {
						var rowExtra = parseInt(missingSpace / localLength);
						for (var j = 0; j < localLength; j++)
							minRows[ localFlexRows[j] ] += rowExtra;
						minRows[lastRow] += missingSpace % localLength;
					} else {
						minRows[lastRow] += missingSpace;
			}}}
		}
	}
}

/*

Children are displayed in a grid.

Layout manager keys:

numCols - Integer. Number of columns in grid.

equalWidth - Boolean. If true, columns are given the same width, and
stretched to fill parent.

hSpacing, vSpacing - Integer. Number of pixels to put between cells.

Default is numCols=1, equalWidth=false, hspacing=vspacing=5

Layout manager hints:

indent - Integer. Number of pixels to indent child from left edge of
cell.

hAlign, vAlign - one of: begin, end, center, fill. Alignment of widget
in cell.

hSpan, vSpan - Integer. Number of cells to span.

hFlex, vFlex - Boolean. If true widget grabs excess space in the
specified direction, i.e. it expands as much as possible.

height, width - Integer. This size will be used on the child, instead of
the layout policy.

Default is hAlign=begin, vAlign=center, indent=0, hSpan=vSpan=1,
hFlex=vFlex=false

*/
